from datetime import date, timedelta
from pathlib import Path

import pytest

from bigquery_etl.backfill.parse import (
    BACKFILL_FILE,
    DEFAULT_REASON,
    DEFAULT_WATCHER,
    Backfill,
    BackfillStatus,
)

DEFAULT_STATUS = BackfillStatus.INITIATE
DEFAULT_BILLING_PROJECT = "moz-fx-data-backfill-slots"

TEST_DIR = Path(__file__).parent.parent

TEST_BACKFILL_1 = Backfill(
    date(2021, 5, 3),
    date(2021, 1, 3),
    date(2021, 5, 3),
    [date(2021, 2, 3)],
    DEFAULT_REASON,
    [DEFAULT_WATCHER],
    DEFAULT_STATUS,
)

TEST_BACKFILL_2 = Backfill(
    date(2023, 5, 3),
    date(2023, 1, 3),
    date(2023, 5, 3),
    [date(2023, 2, 3)],
    DEFAULT_REASON,
    [DEFAULT_WATCHER],
    DEFAULT_STATUS,
)

TEST_BACKFILL_3 = Backfill(
    date(2021, 5, 3),
    date(2021, 1, 3),
    date(2021, 5, 3),
    [date(2021, 2, 3)],
    DEFAULT_REASON,
    [DEFAULT_WATCHER],
    DEFAULT_STATUS,
    "custom_query.sql",
    False,
    False,
    DEFAULT_BILLING_PROJECT,
)


class TestParseBackfill(object):
    def test_backfill_instantiation(self):
        backfill = TEST_BACKFILL_1

        assert backfill.entry_date == date(2021, 5, 3)
        assert backfill.start_date == date(2021, 1, 3)
        assert backfill.end_date == date(2021, 5, 3)
        assert backfill.excluded_dates == [date(2021, 2, 3)]
        assert backfill.reason == DEFAULT_REASON
        assert backfill.watchers == [DEFAULT_WATCHER]
        assert backfill.status == DEFAULT_STATUS
        assert backfill.shredder_mitigation is False
        assert backfill.custom_query_path is None
        assert backfill.billing_project is None
        assert backfill.override_retention_limit is False

    def test_backfill_instantiation_with_billing_project(self):
        backfill = TEST_BACKFILL_3

        assert backfill.entry_date == date(2021, 5, 3)
        assert backfill.start_date == date(2021, 1, 3)
        assert backfill.end_date == date(2021, 5, 3)
        assert backfill.excluded_dates == [date(2021, 2, 3)]
        assert backfill.reason == DEFAULT_REASON
        assert backfill.watchers == [DEFAULT_WATCHER]
        assert backfill.status == DEFAULT_STATUS
        assert backfill.shredder_mitigation is False
        assert backfill.custom_query_path == "custom_query.sql"
        assert backfill.billing_project == DEFAULT_BILLING_PROJECT
        assert backfill.override_retention_limit is False

    def test_invalid_billing_project(self):
        with pytest.raises(ValueError) as e:
            invalid_billing_project = "mozdata"
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
                None,
                None,
                False,
                invalid_billing_project,
            )

        assert "Invalid billing project" in str(e.value)

    def test_invalid_watcher(self):
        with pytest.raises(ValueError) as e:
            invalid_watcher = ["test.org"]
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                invalid_watcher,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid" in str(e.value)
        assert "watchers" in str(e.value)

    def test_invalid_watchers(self):
        with pytest.raises(ValueError) as e:
            invalid_watchers = [DEFAULT_WATCHER, "test.org"]
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                invalid_watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid" in str(e.value)
        assert "watchers" in str(e.value)

    def test_empty_watchers_should_fail(self):
        with pytest.raises(ValueError) as e:
            invalid_watchers = []
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                invalid_watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid" in str(e.value)
        assert "watchers" in str(e.value)

    def test_empty_reason_should_fail(self):
        with pytest.raises(ValueError) as e:
            invalid_reason = None
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                invalid_reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Reason in backfill entry should not be empty" in str(e.value)

    def test_multiple_watchers(self):
        valid_watchers = TEST_BACKFILL_1.watchers + [
            "test2@example.org",
            "test3@example.org",
        ]
        backfill = Backfill(
            TEST_BACKFILL_1.entry_date,
            TEST_BACKFILL_1.start_date,
            TEST_BACKFILL_1.end_date,
            TEST_BACKFILL_1.excluded_dates,
            TEST_BACKFILL_1.reason,
            valid_watchers,
            TEST_BACKFILL_1.status,
        )

        assert backfill.watchers == TEST_BACKFILL_1.watchers + [
            "test2@example.org",
            "test3@example.org",
        ]

    def test_duplicate_watchers_should_fail(self):
        with pytest.raises(ValueError) as e:
            duplicate_watchers = [DEFAULT_WATCHER, DEFAULT_WATCHER]
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                duplicate_watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Duplicate watcher" in str(e.value)

    def test_all_status(self):
        valid_status = [status.value for status in BackfillStatus]
        for i, status in enumerate(BackfillStatus):
            backfill = Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                status,
            )

            assert backfill.status.value == valid_status[i]

    def test_invalid_entry_date_greater_than_today(self):
        with pytest.raises(ValueError) as e:
            invalid_entry_date = date.today() + timedelta(days=1)
            Backfill(
                invalid_entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "can't be in the future" in str(e.value)

    def test_invalid_start_date_greater_than_entry_date(self):
        with pytest.raises(ValueError) as e:
            invalid_start_date = TEST_BACKFILL_1.entry_date + timedelta(days=1)
            Backfill(
                TEST_BACKFILL_1.entry_date,
                invalid_start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid start date" in str(e.value)

    def test_invalid_start_date_greater_than_end_date(self):
        with pytest.raises(ValueError) as e:
            invalid_start_date = TEST_BACKFILL_1.end_date + timedelta(days=1)
            Backfill(
                TEST_BACKFILL_1.entry_date,
                invalid_start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid start date" in str(e.value)

    def test_invalid_end_date_greater_than_entry_date(self):
        with pytest.raises(ValueError) as e:
            invalid_end_date = TEST_BACKFILL_1.entry_date + timedelta(days=1)
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                invalid_end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid end date" in str(e.value)

    def test_invalid_excluded_dates_greater_than_end_date(self):
        with pytest.raises(ValueError) as e:
            invalid_excluded_dates = [TEST_BACKFILL_1.end_date + timedelta(days=1)]
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                invalid_excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid excluded dates" in str(e.value)

    def test_invalid_excluded_dates_greater_than_end_date_multiple(self):
        with pytest.raises(ValueError) as e:
            invalid_excluded_dates = [
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.end_date + timedelta(days=1),
            ]
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                invalid_excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid excluded dates" in str(e.value)

    def test_invalid_excluded_dates_less_than_start_date(self):
        with pytest.raises(ValueError) as e:
            invalid_excluded_dates = [TEST_BACKFILL_1.start_date - timedelta(days=1)]
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                invalid_excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid excluded dates" in str(e.value)

    def test_invalid_excluded_dates_less_than_start_date_multiple(self):
        with pytest.raises(ValueError) as e:
            invalid_excluded_dates = [
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.start_date - timedelta(days=1),
            ]
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                invalid_excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "Invalid excluded dates" in str(e.value)

    def test_excluded_dates_duplicates_should_fail(self):
        with pytest.raises(ValueError) as e:
            invalid_excluded_dates = [date(2021, 2, 3), date(2021, 2, 3)]

            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                invalid_excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "duplicate excluded dates" in str(e.value)

    def test_excluded_dates_not_sorted_should_fail(self):
        with pytest.raises(ValueError) as e:
            invalid_excluded_dates = [date(2021, 2, 4), date(2021, 2, 3)]

            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                invalid_excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                TEST_BACKFILL_1.status,
            )

        assert "excluded dates not sorted" in str(e.value)

    def test_invalid_status(self):
        with pytest.raises(AttributeError):
            invalid_status = "invalid_status"
            Backfill(
                TEST_BACKFILL_1.entry_date,
                TEST_BACKFILL_1.start_date,
                TEST_BACKFILL_1.end_date,
                TEST_BACKFILL_1.excluded_dates,
                TEST_BACKFILL_1.reason,
                TEST_BACKFILL_1.watchers,
                invalid_status,
            )

    def test_non_existing_file(self):
        backfill_file = TEST_DIR / "nonexisting_dir" / BACKFILL_FILE
        with pytest.raises(FileNotFoundError):
            Backfill.entries_from_file(backfill_file)

    def test_of_backfill_file_no_backfill(self):
        backfill_file = TEST_DIR / "test" / BACKFILL_FILE
        with pytest.raises(FileNotFoundError):
            Backfill.entries_from_file(backfill_file)

    def test_of_backfill_file_one(self):
        backfill_file = TEST_DIR / "backfill" / BACKFILL_FILE
        backfills = Backfill.entries_from_file(backfill_file)
        backfill = backfills[0]

        assert backfill.entry_date == date(2021, 5, 3)
        assert backfill.start_date == date(2021, 1, 3)
        assert backfill.end_date == date(2021, 5, 3)
        assert backfill.excluded_dates == [date(2021, 2, 3)]
        assert backfill.reason == DEFAULT_REASON
        assert backfill.watchers == [DEFAULT_WATCHER]
        assert backfill.status == DEFAULT_STATUS

    def test_entries_from_file_multiple(self):
        backfill_file = TEST_DIR / "backfill" / "test_dir_multiple" / BACKFILL_FILE
        backfills = Backfill.entries_from_file(backfill_file)

        backfill_1 = TEST_BACKFILL_1
        backfill_2 = TEST_BACKFILL_2

        assert backfills[0] == backfill_2
        assert backfills[1] == backfill_1

    def test_invalid_file(self):
        backfill_file = TEST_DIR / "test" / "invalid_file_name.yaml"
        with pytest.raises(ValueError) as e:
            Backfill.entries_from_file(backfill_file)

        assert "Invalid file" in str(e.value)

    def test_of_non_existing_table(self):
        backfill_file = TEST_DIR / "non_exist_folder" / BACKFILL_FILE
        with pytest.raises(FileNotFoundError):
            Backfill.entries_from_file(backfill_file)

    def test_is_backfill_file(self):
        assert Backfill.is_backfill_file("foo/bar/invalid.json") is False
        assert Backfill.is_backfill_file("foo/bar/invalid.yaml") is False
        assert Backfill.is_backfill_file(BACKFILL_FILE)
        assert Backfill.is_backfill_file("some/path/to/" + BACKFILL_FILE)

    def test_to_yaml(self):
        expected = (
            "2021-05-03:\n"
            "  start_date: 2021-01-03\n"
            "  end_date: 2021-05-03\n"
            "  excluded_dates:\n"
            "  - 2021-02-03\n"
            "  reason: Please provide a reason for the backfill and links to any related "
            "bugzilla\n"
            "    or jira tickets\n"
            "  watchers:\n"
            "  - nobody@mozilla.com\n"
            "  status: Initiate\n"
            "  shredder_mitigation: false\n"
            "  override_retention_limit: false\n"
        )

        results = TEST_BACKFILL_1.to_yaml()
        assert results == expected

    def test_backfill_str(self):
        backfill = TEST_BACKFILL_1
        actual_backfill_str = str(backfill)
        expected_backfill_str = """
            entry_date = 2021-05-03
            start_date = 2021-01-03
            end_date = 2021-05-03
            excluded_dates = [2021-02-03]
            reason = Please provide a reason for the backfill and links to any related bugzilla or jira tickets
            watcher(s) = [nobody@mozilla.com]
            status = Initiate
            custom_query_path = None
            shredder_mitigation = False
            override_retention_limit = False
            """

        assert actual_backfill_str == expected_backfill_str

    def test_to_yaml_no_excluded_dates(self):
        expected = (
            "2021-05-03:\n"
            "  start_date: 2021-01-03\n"
            "  end_date: 2021-05-03\n"
            "  reason: Please provide a reason for the backfill and links to any related "
            "bugzilla\n"
            "    or jira tickets\n"
            "  watchers:\n"
            "  - nobody@mozilla.com\n"
            "  status: Initiate\n"
            "  shredder_mitigation: false\n"
            "  override_retention_limit: false\n"
        )

        TEST_BACKFILL_1.excluded_dates = []
        results = TEST_BACKFILL_1.to_yaml()

        assert results == expected



"""
Generate user-facing views on top of stable tables.

If there are multiple versions of a given document type, the generated view
references only the most recent.

Note that no view will be generated if a corresponding view is already
present in the target directory, which allows manual overrides of views by
checking them into the sql/ tree of the default branch of the repository.
"""

import logging
from functools import partial
from itertools import groupby
from pathlib import Path

import click
from pathos.multiprocessing import ProcessingPool

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.schema.stable_table_schema import SchemaFile, get_stable_table_schemas
from bigquery_etl.dryrun import get_id_token
from bigquery_etl.config import ConfigLoader

VIEW_QUERY_TEMPLATE = """\
-- Generated via ./bqetl generate stable_views
CREATE OR REPLACE VIEW
  `{full_view_id}`
AS
SELECT
  * REPLACE(
    {replacements}),
  mozfun.norm.extract_version(client_info.app_display_version, 'major') as app_version_major,
  mozfun.norm.extract_version(client_info.app_display_version, 'minor') as app_version_minor,
  mozfun.norm.extract_version(client_info.app_display_version, 'patch') as app_version_patch
FROM
  `{target}`
"""

VIEW_QUERY_TEMPLATE_NO_CLIENT_INFO = """\
-- Generated via ./bqetl generate stable_views
CREATE OR REPLACE VIEW
  `{full_view_id}`
AS
SELECT
  * REPLACE(
    {replacements}),
FROM
  `{target}`
"""

FF_INSTLL_VIEW_QUERY_TEMPLATE = """\
-- Generated via ./bqetl generate stable_views
CREATE OR REPLACE VIEW
    `{full_view_id}`
AS
SELECT
  * REPLACE(
    {replacements}),
  `moz-fx-data-shared-prod`.udf.funnel_derived_installs(
        silent, 
        submission_timestamp, 
        build_id, 
        attribution, 
        distribution_id
    ) AS funnel_derived,
  `moz-fx-data-shared-prod`.udf.distribution_model_installs(distribution_id) AS distribution_model,
  `moz-fx-data-shared-prod`.udf.partner_org_installs(distribution_id) AS partner_org
FROM
  `{target}`
"""

VIEW_METADATA_TEMPLATE = """\
# Generated via ./bqetl generate stable_views
---
friendly_name: Historical Pings for `{document_namespace}/{document_type}`
description: |-
  A historical view of pings sent for the
  `{document_namespace}/{document_type}`
  document type.

  This view is guaranteed to contain only complete days
  (per `submission_timestamp`)
  and to contain only one row per distinct `document_id` within a given date.

  Clustering fields: `normalized_channel`, `sample_id`
"""


def write_dataset_metadata_if_not_exists(
    target_project: str, sql_dir: Path, schema: SchemaFile
):
    """Write default dataset_metadata.yaml files where none exist.

    This function expects to be handed one representative `SchemaFile`
    object representing the dataset.
    """
    from bigquery_etl.metadata.parse_metadata import DatasetMetadata

    dataset_family = schema.bq_dataset_family
    project_dir = sql_dir / target_project

    # Derived dataset
    dataset_name = f"{dataset_family}_derived"
    target = project_dir / dataset_name / "dataset_metadata.yaml"
    target.parent.mkdir(parents=True, exist_ok=True)
    if not target.exists():
        print(f"Creating {target}")
        DatasetMetadata(
            friendly_name=f"{schema.document_namespace} Derived",
            description=(
                f"Derived tables related to document namespace"
                f" {schema.document_namespace},"
                f" usually populated via queries defined in"
                f" https://github.com/mozilla/bigquery-etl"
                f" and managed by Airflow"
            ),
            dataset_base_acl="derived",
            user_facing=False,
        ).write(target)

    # User-facing dataset
    dataset_name = dataset_family
    target = project_dir / dataset_name / "dataset_metadata.yaml"
    target.parent.mkdir(parents=True, exist_ok=True)
    if not target.exists():
        print(f"Creating {target}")
        DatasetMetadata(
            friendly_name=f"{schema.document_namespace}",
            description=(
                f"User-facing views related to document namespace"
                f" {schema.document_namespace}; see https://github.com/"
                f"mozilla-services/mozilla-pipeline-schemas/tree/"
                f"generated-schemas/schemas/{schema.document_namespace}"
            ),
            dataset_base_acl="view",
            user_facing=True,
        ).write(target)


def write_view_if_not_exists(
    target_project: str, sql_dir: Path, id_token=None, schema: SchemaFile = None
):
    """If a view.sql does not already exist, write one to the target directory."""
    # add imports here to run in multiple processes via pathos
    import re

    from bigquery_etl.format_sql.formatter import reformat
    from bigquery_etl.schema import Schema
    from sql_generators.stable_views import (
        FF_INSTLL_VIEW_QUERY_TEMPLATE,
        VIEW_METADATA_TEMPLATE,
        VIEW_QUERY_TEMPLATE,
        VIEW_QUERY_TEMPLATE_NO_CLIENT_INFO,
    )

    VIEW_CREATE_REGEX = re.compile(
        r"CREATE OR REPLACE VIEW\n\s*[^\s]+\s*\nAS", re.IGNORECASE
    )

    SKIP_VIEW_SCHEMA = {
        # skip main and its clones because their large schema causes frequent API
        # failures when trying to update the view schema
        "telemetry.main",
        "telemetry.first_shutdown",
        "telemetry.saved_session",
    }

    target_dir = (
        sql_dir
        / target_project
        / schema.bq_dataset_family
        / schema.bq_table_unversioned
    )

    target_file = target_dir / "view.sql"

    if target_file.exists():
        return

    full_source_id = f"{target_project}.{schema.stable_table}"
    full_view_id = f"{target_project}.{schema.user_facing_view}"
    replacements = ["mozfun.norm.metadata(metadata) AS metadata"]
    key_value_metrics_removed = False
    if schema.schema_id == "moz://mozilla.org/schemas/glean/ping/1":
        replacements += ["mozfun.norm.glean_ping_info(ping_info) AS ping_info"]
        if schema.bq_table == "baseline_v1":
            replacements += [
                "mozfun.norm.glean_baseline_client_info"
                "(client_info, metrics) AS client_info"
            ]
        if (
            schema.bq_dataset_family == "org_mozilla_fenix"
            and schema.bq_table == "metrics_v1"
        ):
            # todo: use mozfun udfs
            metrics_source = (
                "`moz-fx-data-shared-prod`.udf.normalize_fenix_metrics"
                "(client_info.telemetry_sdk_build, metrics)"
            )
        else:
            metrics_source = "metrics"

        datetime_replacements_clause = ""
        metrics_2_aliases = []
        metrics_2_exclusions = []

        if metrics_struct := next(
            (field for field in schema.schema if field["name"] == "metrics"), None
        ):
            if metrics_datetime_fields := [
                metrics_datetime_field["name"]
                for metrics_field in metrics_struct["fields"]
                if metrics_field["name"] == "datetime"
                for metrics_datetime_field in metrics_field["fields"]
            ]:
                datetime_replacements_clause = (
                    f"REPLACE (STRUCT("
                    + ", ".join(
                        field_select
                        for field in metrics_datetime_fields
                        for field_select in (
                            f"mozfun.glean.parse_datetime(metrics.datetime.{field}) AS {field}",
                            f"metrics.datetime.{field} AS raw_{field}",
                        )
                    )
                    + ") AS datetime)"
                )

            # The following metrics were incorrectly deployed as repeated key/value fields and are suffixed with `2`
            # at ingestion to match deployed schemas
            # (see https://github.com/mozilla/gcp-ingestion/blob/9911895cf49ed6364b1b2fb8008310fb60ff9c8e/ingestion-core/src/main/java/com/mozilla/telemetry/ingestion/core/transform/PubsubMessageToObjectNode.java#L376-L384) # noqa E501
            # We're aliasing them to their original names in views
            # (see https://bugzilla.mozilla.org/show_bug.cgi?id=1741487)
            # TODO: Later, after consumers switch to aliased fields we'll remove the `2`-suffixed fields
            metrics_2_types_to_rename = {
                "url2": "url",
                "text2": "text",
                "jwe2": "jwe",
                "labeled_rate2": "labeled_rate",
            }

            # Due to past error in the deployment all of the following fields are considered invalid (`url`, `text`, `jwe`, `labeled_rate`)
            # after the issue has been resolved these fields were corrected and have the suffix `2`.
            # This suffix is automatically added by the ingestion pipeline and is a hidden detail from the data producers.
            # For this reason, all of the mentioned fields without the suffix can be excluded from the view
            # and any of the suffixed fields can have the suffix removed to make the data available more consistent with how it is provided.
            #
            # We have to handle these fields in two stages via `EXCEPT` and aliases instead of
            # a single `REPLACE` because there are some tables that have `url2` field but not `url` field.
            for metrics_field in metrics_struct["fields"]:
                # Excluding non-suffixed fields are they are considered invalid
                if metrics_field["name"] in metrics_2_types_to_rename.values():
                    metrics_2_exclusions += [metrics_field["name"]]
                    key_value_metrics_removed = True

                # Using aliasing to remove the suffix from the valid fields
                if metrics_field["name"] in metrics_2_types_to_rename:
                    metrics_2_aliases += [
                        f"metrics.{metrics_field['name']} AS {metrics_2_types_to_rename[metrics_field['name']]}"
                    ]

        if datetime_replacements_clause or metrics_2_aliases or metrics_2_exclusions:
            except_clause = ""
            if metrics_2_exclusions:
                except_clause = "EXCEPT (" + ", ".join(metrics_2_exclusions) + ")"
            metrics_select = (
                f"{metrics_source}.* {except_clause} {datetime_replacements_clause}"
            )

            replacements += [
                f"(SELECT AS STRUCT "
                + ", ".join([metrics_select] + metrics_2_aliases)
                + ") AS metrics"
            ]
        elif metrics_source != "metrics":
            replacements += [f"{metrics_source} AS metrics"]
        if schema.bq_dataset_family == "firefox_desktop":
            # FOG does not provide an app_name, so we inject the one that
            # people already associate with desktop Firefox per bug 1672191.
            replacements += [
                "'Firefox' AS normalized_app_name",
            ]
    elif (
        schema.schema_id.startswith("moz://mozilla.org/schemas/main/ping/")
        and "_use_counter_" not in schema.stable_table
    ):
        replacements += [
            "`moz-fx-data-shared-prod`.udf.normalize_main_payload(payload) AS payload"
        ]
    replacements_str = ",\n    ".join(replacements)

    # For the firefox_installer.install view, use the FF_INSTLL_VIEW_QUERY_TEMPLATE template
    if full_view_id == "moz-fx-data-shared-prod.firefox_installer.install":
        full_sql = reformat(
            FF_INSTLL_VIEW_QUERY_TEMPLATE.format(
                target=full_source_id,
                replacements=replacements_str,
                full_view_id=full_view_id,
            ),
            trailing_newline=True,
        )
    # If it's a glean stable view, include the app version parsing columns
    elif schema.schema_id == "moz://mozilla.org/schemas/glean/ping/1":
        full_sql = reformat(
            VIEW_QUERY_TEMPLATE.format(
                target=full_source_id,
                replacements=replacements_str,
                full_view_id=full_view_id,
            ),
            trailing_newline=True,
        )
    # For all other views, don't include the app version parsing columns
    else:
        full_sql = reformat(
            VIEW_QUERY_TEMPLATE_NO_CLIENT_INFO.format(
                target=full_source_id,
                replacements=replacements_str,
                full_view_id=full_view_id,
            ),
            trailing_newline=True,
        )
    print(f"Creating {target_file}")
    target_dir.mkdir(parents=True, exist_ok=True)
    with target_file.open("w") as f:
        f.write(full_sql)
    metadata_content = VIEW_METADATA_TEMPLATE.format(
        document_namespace=schema.document_namespace,
        document_type=schema.document_type,
    )
    metadata_file = target_dir / "metadata.yaml"
    if not metadata_file.exists():
        with metadata_file.open("w") as f:
            f.write(metadata_content)

    if schema.user_facing_view not in SKIP_VIEW_SCHEMA:
        # get view schema with descriptions
        try:
            content = VIEW_CREATE_REGEX.sub("", target_file.read_text())
            content += " WHERE DATE(submission_timestamp) = '2020-01-01'"
            view_schema = Schema.from_query_file(
                target_file, content=content, sql_dir=sql_dir, id_token=id_token
            )

            stable_table_schema = Schema.from_json({"fields": schema.schema})

            # This is needed if we removed any of the `url`, `text`, `jwe`, or `labeled_rate`
            # from the view schema since these fields exist in the source table
            ignore_missing_fields = key_value_metrics_removed

            view_schema.merge(
                stable_table_schema,
                attributes=["description"],
                add_missing_fields=False,
                ignore_missing_fields=ignore_missing_fields,
            )
            view_schema.to_yaml_file(target_dir / "schema.yaml")
        except Exception as e:
            print(f"Cannot generate schema.yaml for {target_file}: {e}")


@click.command("generate")
@click.option(
    "--target_project",
    "--target-project",
    help="Which project the views should be written to.",
    default="moz-fx-data-shared-prod",
)
@click.option(
    "--output-dir",
    "--output_dir",
    "--sql-dir",
    "--sql_dir",
    help="The location to write to. Defaults to sql/.",
    default=Path("sql"),
    type=click.Path(file_okay=False),
)
@click.option(
    "--log-level",
    "--log_level",
    help="Log level.",
    default=logging.getLevelName(logging.INFO),
    type=str.upper,
)
@click.option(
    "--parallelism",
    "-P",
    help="Maximum number of tasks to execute concurrently",
    default=20,
    type=int,
)
@use_cloud_function_option
def generate(target_project, output_dir, log_level, parallelism, use_cloud_function):
    """
    Generate view definitions.

    Metadata about document types is from the generated-schemas branch
    of mozilla-pipeline-schemas. We write out generated views for each
    document type in parallel.

    There is a performance bottleneck here due to the need to dry-run each
    view to ensure the source table actually exists.
    """
    # set log level
    logging.basicConfig(level=log_level, format="%(levelname)s %(message)s")

    schemas = get_stable_table_schemas()
    one_schema_per_dataset = [
        last
        for k, (*_, last) in groupby(schemas, lambda t: t.bq_dataset_family)
        if k
        not in ConfigLoader.get("generate", "stable_views", "skip_datasets", fallback=[])
    ]

    id_token = get_id_token()

    with ProcessingPool(parallelism) as pool:
        pool.map(
            partial(
                write_view_if_not_exists, target_project, Path(output_dir), id_token
            ),
            schemas,
        )
        pool.map(
            partial(
                write_dataset_metadata_if_not_exists,
                target_project,
                Path(output_dir),
            ),
            one_schema_per_dataset,
        )



"""Validate metadata files."""

import logging
import os
import sys
from argparse import ArgumentParser
from pathlib import Path

import click
import yaml

from bigquery_etl.config import ConfigLoader
from bigquery_etl.schema import SCHEMA_FILE, Schema

from ..util import standard_args
from ..util.common import extract_last_group_by_from_query, project_dirs
from .parse_metadata import DatasetMetadata, Metadata

parser = ArgumentParser(description=__doc__)

parser.add_argument("--target", help="File or directory containing metadata files")
standard_args.add_log_level(parser)

CODEOWNERS_FILE = "CODEOWNERS"
CHANGE_CONTROL_LABEL = "change_controlled"
SHREDDER_MITIGATION_LABEL = "shredder_mitigation"
ID_LEVEL_COLUMNS_FILE_PATH = Path(__file__).parent / "id_level_columns.yaml"


def validate_public_data(metadata, path):
    """Check if the metadata for public data queries is valid."""
    is_valid = True

    if metadata.is_public_bigquery() or metadata.is_public_json():
        if not metadata.review_bugs():
            logging.error(f"Missing review bug for public data: {path}")
            is_valid = False

    return is_valid


def validate_change_control(
    file_path,
    metadata,
    codeowners_file,
    project_id=ConfigLoader.get(
        "default", "project", fallback="moz-fx-data-shared-prod"
    ),
    sql_dir=ConfigLoader.get("default", "sql_dir", fallback="sql"),
):
    """Verify that a query is correctly setup for change control."""
    path_to_add = file_path.partition(f"{project_id}/")[2]
    path_in_codeowners = os.path.join(sql_dir, project_id, path_to_add)
    has_change_control = CHANGE_CONTROL_LABEL in metadata.labels

    if has_change_control:
        # This label requires to have at least one owner in the metadata file.
        # And for any of the owners, at least one entry in the CODEOWNERS file.
        # The owners can be emails or GitHub identities e.g. mozilla/team_name.
        if len(metadata.owners) == 0:
            click.echo(
                click.style(
                    f"The metadata for {file_path} has the label"
                    f" change_controlled but it's missing owners."
                )
            )
            return False

        with open(codeowners_file, "r") as owners_file:
            file_content = owners_file.readlines()
            content = [line for line in file_content if not line.startswith("#")]

        owners_list = []
        for owner in metadata.owners:
            if "@" not in owner:
                owner = f"@{owner}"
            owners_list.append(owner)
        sample_row_all_owners = f"/{path_in_codeowners} {(' '.join(owners_list))}"

        if not [line for line in content if path_in_codeowners in line]:
            click.echo(
                click.style(
                    f"ERROR: This query has label `change_controlled` which "
                    f"requires the query path and at least one of the owners in "
                    f"CODEOWNERS. Sample row expected: {sample_row_all_owners}"
                )
            )
            return False

        for line in content:
            if path_in_codeowners in line and not any(
                owner in line for owner in owners_list
            ):
                click.echo(
                    click.style(
                        f"ERROR: This query has label `change_controlled` which "
                        f"requires at least one of the owners to be registered "
                        f"in CODEOWNERS. Sample row expected: \n"
                        f"{sample_row_all_owners}"
                    )
                )
                return False
    return True


def validate_shredder_mitigation(query_dir, metadata):
    """Check queries with shredder mitigation label comply with requirements."""
    has_shredder_mitigation = SHREDDER_MITIGATION_LABEL in metadata.labels

    if has_shredder_mitigation:
        schema_file = Path(query_dir) / SCHEMA_FILE
        schema = Schema.from_schema_file(schema_file).to_bigquery_schema()

        # This label requires that the query doesn't have id-level columns,
        # has a group by that is explicit & all schema columns have descriptions.
        query_file = Path(query_dir) / "query.sql"
        query_group_by = extract_last_group_by_from_query(sql_path=query_file)

        # Validate that the query group by is as required.
        integers_in_group_by = False
        for e in query_group_by:
            try:
                int(e)
                integers_in_group_by = True
            except ValueError:
                continue
        if (
            "ALL" in query_group_by
            or not all(isinstance(e, str) for e in query_group_by)
            or not query_group_by
            or integers_in_group_by
        ):
            click.echo(
                "Shredder mitigation validation failed, GROUP BY must use an explicit list "
                "of columns. Avoid expressions like `GROUP BY ALL` or `GROUP BY 1, 2, 3`."
            )
            return False

        with open(ID_LEVEL_COLUMNS_FILE_PATH, "r") as columns_file:
            columns_from_file = yaml.safe_load(columns_file)
            id_level_columns = columns_from_file.get("id_level_columns", [])

        for field in schema:
            # Validate that the query columns have descriptions.
            if not field.description:
                click.echo(
                    f"Shredder mitigation validation failed, {field.name} does not have "
                    f"a description in the schema."
                )
                return False
            # Validate that id-level columns are not present in the query schema.
            if field.name in id_level_columns:
                click.echo(
                    f"Shredder mitigation validation failed, {field.name} is an id-level"
                    f" column that is not allowed for this type of backfill."
                )
                return False
    return True


def validate_deprecation(metadata, path):
    """Check that deprecated is True when deletion date exists."""
    if metadata.deletion_date and not metadata.deprecated:
        click.echo(
            f"Deletion date should only be added when table is deprecated in {path}"
        )
        return False

    return True


def validate_exclusion_list_expiration_days(metadata, path):
    """Check if any of the retention exclusion tables have expiration_days set."""
    is_valid = True
    retention_exclusion_list = set(
        ConfigLoader.get("retention_exclusion_list", fallback=[])
    )
    normalized_path = str(Path(path).parent)

    if normalized_path in retention_exclusion_list:
        # Access and check expiration_days metadata
        expiration_days = (
            metadata.bigquery.time_partitioning.expiration_days
            if metadata
            and getattr(metadata, "bigquery", None)
            and getattr(metadata.bigquery, "time_partitioning", None)
            else None
        )

        if expiration_days is not None:
            click.echo(
                click.style(
                    f"ERROR: Table at {path} is in the retention exclusion list but has expiration_days set to {expiration_days}.",
                    fg="red",
                )
            )
            is_valid = False

    return is_valid


def validate_retention_policy_based_on_table_type(metadata, path):
    """Check if any of the retention exclusion tables have expiration_days set."""
    is_valid = True
    table_type = (
        metadata.labels.get("table_type") if isinstance(metadata.labels, dict) else None
    )
    expiration_days = (
        metadata.bigquery.time_partitioning.expiration_days
        if metadata
        and getattr(metadata, "bigquery", None)
        and getattr(metadata.bigquery, "time_partitioning", None)
        else None
    )

    # retention_exclusion_list = set(
    #     ConfigLoader.get("retention_exclusion_list", fallback=[])
    # )

    # normalized_path = str(Path(path).parent)
    if expiration_days is not None and table_type == "aggregate":
        click.echo(
            click.style(
                f"ERROR: Table at {path} is an aggregate table but has expiration_days set to {expiration_days}.",
                fg="red",
            )
        )
        is_valid = False
    # The below line of code should be uncommented when the retention project is completed
    # if (
    #     expiration_days is None
    #     and table_type == "client_level"
    #     and normalized_path not in retention_exclusion_list
    # ):
    #     click.echo(
    #         click.style(
    #             f"ERROR: Table at {path} is an client level table and needs expiration_days to be set",
    #             fg="red",
    #         )
    #     )
    #     is_valid = False
    return is_valid


def validate(target):
    """Validate metadata files."""
    failed = False

    if os.path.isdir(target):
        for root, dirs, files in os.walk(target, followlinks=True):
            for file in files:
                if Metadata.is_metadata_file(file):
                    path = os.path.join(root, file)
                    metadata = Metadata.from_file(path)

                    if not validate_public_data(metadata, path):
                        failed = True

                    if not validate_change_control(
                        file_path=root,
                        metadata=metadata,
                        codeowners_file=CODEOWNERS_FILE,
                    ):
                        failed = True

                    if not validate_shredder_mitigation(
                        query_dir=root,
                        metadata=metadata,
                    ):
                        failed = True

                    if not validate_deprecation(metadata, path):
                        failed = True

                    if not validate_exclusion_list_expiration_days(metadata, path):
                        failed = True

                    if not validate_retention_policy_based_on_table_type(
                        metadata, path
                    ):
                        failed = True

                    # todo more validation
                    # e.g. https://github.com/mozilla/bigquery-etl/issues/924
    else:
        logging.error(f"Invalid target: {target}, target must be a directory.")
        sys.exit(1)

    if failed:
        sys.exit(1)


def validate_datasets(target):
    """Validate dataset metadata files."""
    failed = False

    if os.path.isdir(target):
        for root, dirs, files in os.walk(target, followlinks=True):
            for file in files:
                if DatasetMetadata.is_dataset_metadata_file(file):
                    path = os.path.join(root, file)
                    _ = DatasetMetadata.from_file(path)
    else:
        logging.error(f"Invalid target: {target}, target must be a directory.")
        sys.exit(1)

    if failed:
        sys.exit(1)


def main():
    """Validate all metadata.yaml files in the provided target directory."""
    args = parser.parse_args()

    # set log level
    try:
        logging.basicConfig(level=args.log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        parser.error(f"argument --log-level: {e}")

    if args.target:
        validate(args.target)
    else:
        for project in project_dirs():
            validate(project)


if __name__ == "__main__":
    main()



"""bigquery-etl CLI check command."""

import re
import subprocess
import sys
import tempfile
from pathlib import Path
from subprocess import CalledProcessError
from typing import List, Optional, Union

import click
import sqlparse

from bigquery_etl.format_sql.formatter import reformat

from ..cli.utils import (
    is_authenticated,
    paths_matching_checks_pattern,
    project_id_option,
    sql_dir_option,
)
from ..util.common import render as render_template

DEFAULT_MARKER = "fail"


def _build_jinja_parameters(query_args):
    """Convert the bqetl parameters to a dictionary for use by the Jinja template."""
    parameters = {}
    for query_arg in query_args:
        param_and_value = query_arg.split("=")
        if len(param_and_value) == 2:
            # e.g. --parameter=download_date:DATE:2023-05-28
            # the dict result is {"download_date": "2023-05-28"}
            bq_parameter = param_and_value[1].split(":")
            if len(bq_parameter) == 3:
                if re.match(r"^\w+$", bq_parameter[0]):
                    parameters[bq_parameter[0]] = bq_parameter[2]
            else:
                # e.g. --project_id=moz-fx-data-marketing-prod
                # the dict result is {"project_id": "moz-fx-data-marketing-prod"}
                if param_and_value[0].startswith("--"):
                    parameters[param_and_value[0].strip("--")] = param_and_value[1]
        else:
            if query_arg == "--dry_run":
                continue

            print(f"parameter {query_arg} will not be used to render Jinja template.")
    return parameters


def _render_result_split_by_marker(marker, rendered_result):
    """Filter the rendered sql checks with the set marker."""
    extracted_result = []
    rendered_result = sqlparse.split(rendered_result)

    for sql_statement in rendered_result:
        sql_statement = sql_statement.strip()
        if re.search(f"#{marker}", sql_statement, re.IGNORECASE):
            extracted_result.append(sql_statement)
    return " ".join(extracted_result)


def _parse_check_output(output: str) -> str:
    output = output.replace("\n", " ")
    if "ETL Data Check Failed:" in output:
        return f"ETL Data Check Failed:{output.split('ETL Data Check Failed:')[1]}"
    return output


@click.group(
    help="""
        Commands for managing and running bqetl data checks.
    """
)
@click.pass_context
def check(ctx):
    """Create the CLI group for the check command."""
    # create temporary directory generated content is written to
    # the directory will be deleted automatically after the command exits
    ctx.ensure_object(dict)
    ctx.obj["TMP_DIR"] = ctx.with_resource(tempfile.TemporaryDirectory())


@check.command(
    help="""
    Renders data check query using parameters provided (OPTIONAL).
    \b
    Returns the raw SQL query of the conditions specified in the corresponding checks.sql file

    Example:

    \t./bqetl check render --project_id=moz-fx-data-marketing-prod ga_derived.downloads_with_attribution_v2 --parameter=download_date:DATE:2023-05-01
    """,
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("dataset")
@project_id_option()
@sql_dir_option
@click.pass_context
def render(
    ctx: click.Context, dataset: str, project_id: Optional[str], sql_dir: Optional[str]
) -> None:
    """Render a check's Jinja template."""
    for checks_file, project_id, dataset_id, table in paths_matching_checks_pattern(
        dataset, sql_dir, project_id=project_id
    ):
        click.echo(
            _render(
                checks_file,
                dataset_id,
                table,
                project_id=project_id,
                query_arguments=ctx.args[:],
            )
        )

    return None


def _render(
    checks_file: Path,
    dataset_id: str,
    table: str,
    project_id: Union[str, None] = None,
    query_arguments: List[str] = list(),
):
    if checks_file is None:
        return

    checks_file = Path(checks_file)

    query_arguments.append("--use_legacy_sql=false")

    if project_id is not None:
        query_arguments.append(f"--project_id={project_id}")

    # Convert all the Airflow params to jinja usable dict.
    parameters = _build_jinja_parameters(query_arguments)

    jinja_params = {
        **{"dataset_id": dataset_id, "table_name": table},
        **parameters,
    }

    rendered_check_query = render_template(
        checks_file.name,
        template_folder=str(checks_file.parent),
        templates_dir="",
        format=False,
        **jinja_params,
    )

    # replace query @params with param values passed via the cli
    for param, value in parameters.items():
        if param in rendered_check_query:
            rendered_check_query = rendered_check_query.replace(
                f"@{param}", f'"{value}"'
            )

    rendered_check_query = reformat(rendered_check_query)

    return rendered_check_query


@check.command(
    help="""
    Runs data checks defined for the query (checks.sql).

    Example:

    \t./bqetl check run ga_derived.downloads_with_attribution_v2 --parameter=download_date:DATE:2023-05-01

    Checks can be validated using the `--dry_run` flag without executing them:

    \t./bqetl check run --dry_run ga_derived.downloads_with_attribution_v2 --parameter=download_date:DATE:2023-05-01
    """,
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("dataset")
@project_id_option()
@sql_dir_option
@click.option("--marker", default=DEFAULT_MARKER, help="Marker to filter checks.")
@click.option(
    "--dry_run",
    "--dry-run",
    is_flag=True,
    default=False,
    help="To dry run the query to make sure it is valid",
)
@click.pass_context
def run(ctx, dataset, project_id, sql_dir, marker, dry_run):
    """Run a check."""
    if not is_authenticated():
        click.echo(
            "Authentication to GCP required. Run `gcloud auth login  --update-adc` "
            "and check that the project is set correctly."
        )
        sys.exit(1)

    for checks_file, project_id, dataset_id, table in paths_matching_checks_pattern(
        dataset, sql_dir, project_id=project_id
    ):
        _run_check(
            checks_file,
            project_id,
            dataset_id,
            table,
            ctx.args,
            dry_run=dry_run,
            marker=marker,
        )


def _run_check(
    checks_file,
    project_id,
    dataset_id,
    table,
    query_arguments,
    marker=DEFAULT_MARKER,
    dry_run=False,
):
    """Run the check."""
    if checks_file is None:
        return

    checks_file = Path(checks_file)

    query_arguments.append("--use_legacy_sql=false")
    if project_id is not None:
        query_arguments.append(f"--project_id={project_id}")

    if dry_run is True:
        query_arguments.append("--dry_run")

    # Convert all the Airflow params to jinja usable dict.
    parameters = _build_jinja_parameters(query_arguments)

    jinja_params = {
        **{"dataset_id": dataset_id, "table_name": table},
        **parameters,
    }
    if "format" not in jinja_params:
        jinja_params["format"] = False

    rendered_result = render_template(
        checks_file.name,
        template_folder=str(checks_file.parent),
        templates_dir="",
        **jinja_params,
    )
    result_split_by_marker = _render_result_split_by_marker(marker, rendered_result)
    checks = sqlparse.split(result_split_by_marker)
    seek_location = 0
    check_failed = False

    with tempfile.NamedTemporaryFile(mode="w+") as query_stream:
        for rendered_check in checks:
            # since the last check will end with ; the last entry will be empty string.
            if len(rendered_check) == 0:
                continue
            rendered_check = rendered_check.strip()
            query_stream.write(rendered_check)
            query_stream.seek(seek_location)
            seek_location += len(rendered_check)

            # run the query as shell command so that passed parameters can be used as is
            try:
                subprocess.check_output(
                    ["bq", "query"] + query_arguments,
                    stdin=query_stream,
                    encoding="UTF-8",
                )
            except CalledProcessError as e:
                print(_parse_check_output(e.output))
                check_failed = True

    if check_failed:
        sys.exit(1)


# todo: add validate method -- there must always be #fail checks



"""Use counter table generation."""

import os
import shutil
from pathlib import Path

import click
import yaml
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.common import write_sql

FILE_PATH = Path(os.path.dirname(__file__))
BASE_DIR = Path(FILE_PATH).parent.parent
TEMPLATE_CONFIG = FILE_PATH / "templates/templating.yaml"


def generate_query(
    project: str,
    dataset: str,
    destination_table: str,
    write_dir: Path,
    query_fpath: str,
):
    """Generate use counter table query."""
    with open(TEMPLATE_CONFIG, "r") as f:
        render_kwargs = yaml.safe_load(f) or {}
    env = Environment(loader=FileSystemLoader(str(FILE_PATH / "templates")))
    template = env.get_template(query_fpath)

    write_sql(
        write_dir / project,
        f"{project}.{dataset}.{destination_table}",
        "query.sql",
        reformat(template.render(**render_kwargs)),
    )


def generate_view(project: str, dataset: str, destination_table: str, write_dir: Path):
    """Generate feature usage table view."""
    view_name = destination_table.split("_v")[0]
    view_dataset = dataset.split("_derived")[0]

    sql = reformat(
        f"""
        CREATE OR REPLACE VIEW `{project}.{view_dataset}.{view_name}` AS
        SELECT
            *
        FROM
            `mozilla-public-data.{dataset}.{destination_table}`
    """
    )

    write_sql(
        write_dir / project, f"{project}.{view_dataset}.{view_name}", "view.sql", sql
    )


def generate_metadata(
    project: str,
    dataset: str,
    destination_table: str,
    write_dir: Path,
    friendly_table_name: str,
):
    """Copy metadata.yaml file to destination directory."""
    env = Environment(loader=FileSystemLoader(str(FILE_PATH / "templates")))
    template = env.get_template("metadata.yaml")

    path_to_write_metadata = (
        write_dir / project / dataset / destination_table / "metadata.yaml"
    )
    path_to_write_metadata.write_text(
        template.render({"friendly_table_name": friendly_table_name})
    )


def generate_schema(
    project: str, dataset: str, destination_table: str, write_dir: Path
):
    """Generate schema"""
    shutil.copyfile(
        FILE_PATH / "templates" / "schema.yaml",
        write_dir / project / dataset / destination_table / "schema.yaml",
    )


@click.command("generate")
@click.option(
    "--target-project",
    "--target_project",
    help="Which project the queries should be written to.",
    default="moz-fx-data-shared-prod",
)
@click.option(
    "--output-dir",
    "--output_dir",
    help="The location to write to. Defaults to sql/.",
    default=Path("sql"),
    type=click.Path(file_okay=False),
)
@use_cloud_function_option
@click.pass_context
def generate(ctx, target_project, output_dir, use_cloud_function):
    """Generate the use counter tables for Fenix and Firefox Desktop"""
    output_dir = Path(output_dir)

    config = {
        "fenix": {
            "query_fpath": "query_fenix.sql",
            "dataset": "fenix_derived",
            "dst_tbl": "fenix_use_counters_v2",
            "friendly_nm": "Fenix Use Counters V2",
        },
        "firefox": {
            "query_fpath": "query_ff_desktop.sql",
            "dataset": "firefox_desktop_derived",
            "dst_tbl": "firefox_desktop_use_counters_v2",
            "friendly_nm": "Firefox Desktop Use Counters V2",
        },
    }

    # FIRST - generate Fenix one
    generate_query(
        target_project,
        dataset=config["fenix"]["dataset"],
        destination_table=config["fenix"]["dst_tbl"],
        write_dir=output_dir,
        query_fpath=config["fenix"]["query_fpath"],
    )
    generate_view(
        target_project,
        dataset=config["fenix"]["dataset"],
        destination_table=config["fenix"]["dst_tbl"],
        write_dir=output_dir,
    )
    generate_metadata(
        target_project,
        dataset=config["fenix"]["dataset"],
        destination_table=config["fenix"]["dst_tbl"],
        write_dir=output_dir,
        friendly_table_name=config["fenix"]["friendly_nm"],
    )
    generate_schema(
        target_project,
        dataset=config["fenix"]["dataset"],
        destination_table=config["fenix"]["dst_tbl"],
        write_dir=output_dir,
    )
    # NEXT - generate firefox desktop one
    generate_query(
        project=target_project,
        dataset=config["firefox"]["dataset"],
        destination_table=config["firefox"]["dst_tbl"],
        write_dir=output_dir,
        query_fpath=config["firefox"]["query_fpath"],
    )
    generate_view(
        project=target_project,
        dataset=config["firefox"]["dataset"],
        destination_table=config["firefox"]["dst_tbl"],
        write_dir=output_dir,
    )
    generate_metadata(
        project=target_project,
        dataset=config["firefox"]["dataset"],
        destination_table=config["firefox"]["dst_tbl"],
        write_dir=output_dir,
        friendly_table_name=config["firefox"]["friendly_nm"],
    )
    generate_schema(
        project=target_project,
        dataset=config["firefox"]["dataset"],
        destination_table=config["firefox"]["dst_tbl"],
        write_dir=output_dir,
    )



#!/usr/bin/env python3

"""Determine column sizes by performing dry runs."""

from argparse import ArgumentParser
from fnmatch import fnmatchcase
from functools import partial
from multiprocessing.pool import ThreadPool

from google.cloud import bigquery

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--dataset", default="*_stable")
parser.add_argument("--destination_dataset", default="monitoring_derived")
parser.add_argument("--destination_table", default="column_size_v1")


def get_columns(client, project, dataset):
    """Return list of all columns in each table of a dataset."""
    sql = (
        f"SELECT table_name, field_path AS column_name "
        f"FROM `{project}.{dataset}.INFORMATION_SCHEMA.COLUMN_FIELD_PATHS` "
    )

    try:
        result = client.query(sql).result()
        return [(dataset, row.table_name, row.column_name) for row in result]
    except Exception as e:
        print(f"Error querying dataset {dataset}")

    return []


def get_column_size_json(client, date, column):
    """Return the size of a specific date partition of the specified table."""
    try:
        job_config = bigquery.QueryJobConfig(dry_run=True, use_query_cache=False)
        dataset_id, table_id, column_name = column

        sql = f"""
            SELECT {column_name} FROM {dataset_id}.{table_id}
            WHERE DATE(submission_timestamp) = '{date}'
        """

        job = client.query(sql, job_config=job_config)

        size = job.total_bytes_processed

        return {
            "submission_date": date,
            "dataset_id": dataset_id,
            "table_id": table_id,
            "column_name": column_name,
            "byte_size": size,
        }
    except Exception as e:
        return None


def save_column_sizes(
    client, column_sizes, date, destination_dataset, destination_table
):
    """Write column sizes for tables for a specific day to BigQuery."""
    job_config = bigquery.LoadJobConfig()
    job_config.schema = (
        bigquery.SchemaField("submission_date", "DATE"),
        bigquery.SchemaField("dataset_id", "STRING"),
        bigquery.SchemaField("table_id", "STRING"),
        bigquery.SchemaField("column_name", "STRING"),
        bigquery.SchemaField("byte_size", "INT64"),
    )
    job_config.write_disposition = bigquery.job.WriteDisposition.WRITE_TRUNCATE

    partition_date = date.replace("-", "")
    client.load_table_from_json(
        column_sizes,
        f"{destination_dataset}.{destination_table}${partition_date}",
        job_config=job_config,
    ).result()


def main():
    """Entrypoint for the column size job."""
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    datasets = [
        dataset.dataset_id
        for dataset in list(client.list_datasets())
        if fnmatchcase(dataset.dataset_id, args.dataset)
    ]

    with ThreadPool(20) as p:
        table_columns = [
            column
            for dataset in datasets
            for column in get_columns(client, args.project, dataset)
        ]

        column_sizes = p.map(
            partial(get_column_size_json, client, args.date),
            table_columns,
            chunksize=1,
        )

        column_sizes = [cs for cs in column_sizes if cs is not None]

        save_column_sizes(
            client,
            column_sizes,
            args.date,
            args.destination_dataset,
            args.destination_table,
        )


if __name__ == "__main__":
    main()



"""Generate and run clients_last_seen_joined queries for Glean apps."""

from sql_generators.glean_usage.common import GleanTable

TARGET_TABLE_ID = "clients_last_seen_joined_v1"
PREFIX = "clients_last_seen_joined"


class ClientsLastSeenJoined(GleanTable):
    """Represents generated clients_last_seen_joined table."""

    def __init__(self):
        """Initialize clients_last_seen_joined table."""
        GleanTable.__init__(self)
        self.target_table_id = TARGET_TABLE_ID
        self.per_app_id_enabled = False
        self.cross_channel_template = None



from pathlib import Path

import pytest

from bigquery_etl.metadata.parse_metadata import (
    DatasetMetadata,
    Metadata,
    PartitionType,
)

TEST_DIR = Path(__file__).parent.parent


class TestParseMetadata(object):
    def test_metadata_instantiation(self):
        metadata = Metadata(
            "Test metadata", "test description", ["test@example.org"], {}
        )

        assert metadata.friendly_name == "Test metadata"
        assert metadata.description == "test description"
        assert metadata.owners == ["test@example.org"]
        assert metadata.labels == {}
        assert metadata.scheduling == {}

    def test_invalid_owners(self):
        with pytest.raises(ValueError):
            Metadata("Test metadata", "test description", ["testexample.org"])

    def test_invalid_label(self):
        with pytest.raises(ValueError):
            Metadata(
                "Test metadata",
                "test description",
                ["test@example.org"],
                {"INVALID-KEY": "foo"},
            )
        with pytest.raises(ValueError):
            Metadata(
                "Test metadata",
                "test description",
                ["test@example.org"],
                {"foo": "INVALID-VALUE"},
            )

    def test_invalid_review_bugs(self):
        assert Metadata(
            "Test", "Description", ["test@test.org"], {"review_bugs": [123456]}
        )
        with pytest.raises(ValueError):
            Metadata(
                "Test",
                "Description",
                ["test@example.org"],
                {"review_bugs": 123456},
            )

    def test_is_valid_label(self):
        assert Metadata.is_valid_label("valid_label")
        assert Metadata.is_valid_label("valid-label1")
        assert Metadata.is_valid_label("1231")
        assert Metadata.is_valid_label("1231-21")
        assert Metadata.is_valid_label("a" * 63)
        assert Metadata.is_valid_label("låbel") is False
        assert Metadata.is_valid_label("a" * 64) is False
        assert Metadata.is_valid_label("INVALID") is False
        assert Metadata.is_valid_label("invalid.label") is False
        assert Metadata.is_valid_label("") is False

    def test_from_file(self):
        metadata_file = TEST_DIR / "data" / "metadata.yaml"
        metadata = Metadata.from_file(metadata_file)

        assert metadata.friendly_name == "Test metadata file"
        assert metadata.description == "Please provide a description for the query"
        assert "schedule" in metadata.labels
        assert metadata.labels["schedule"] == "daily"
        assert "public_json" in metadata.labels
        assert metadata.labels["public_json"] == ""
        assert metadata.is_public_json()
        assert metadata.is_incremental()
        assert metadata.is_incremental_export()
        assert metadata.review_bugs() is None
        assert "1232341234" in metadata.labels
        assert "1234_abcd" in metadata.labels
        assert "number_value" in metadata.labels
        assert metadata.labels["number_value"] == "1234234"
        assert "number_string" in metadata.labels
        assert metadata.labels["number_string"] == "1234abcde"
        assert "123-432" in metadata.labels
        assert metadata.owners == [
            "test1@mozilla.com",
            "test2@example.com",
            "mozilla/test_group",
        ]
        assert metadata.labels["owner1"] == "test1"
        assert metadata.labels["owner2"] == "test2"
        assert "query.sql" in metadata.references
        assert metadata.references["query.sql"] == ["project.dataset_derived.table_v1"]

    def test_non_existing_file(self):
        metadata_file = TEST_DIR / "nonexisting_dir" / "metadata.yaml"
        with pytest.raises(FileNotFoundError):
            Metadata.from_file(metadata_file)

    def test_of_query_file(self):
        metadata_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "non_incremental_query_v1"
            / "query.sql"
        )
        metadata = Metadata.of_query_file(metadata_file)

        assert metadata.friendly_name == "Test table for a non-incremental query"
        assert metadata.description == "Test table for a non-incremental query"
        assert metadata.review_bugs() == ["1999999", "12121212"]
        assert metadata.workgroup_access is not None
        assert metadata.workgroup_access[0].role == "roles/bigquery.dataViewer"
        assert (
            metadata.workgroup_access[0].members[0] == "workgroup:dataops-managed/taar"
        )

    def test_of_query_file_no_metadata(self):
        metadata_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "no_metadata_query_v1"
            / "query.sql"
        )
        with pytest.raises(FileNotFoundError):
            Metadata.of_query_file(metadata_file)

    def test_of_table(self):
        metadata = Metadata.of_table(
            "test",
            "non_incremental_query",
            "v1",
            TEST_DIR / "data" / "test_sql" / "moz-fx-data-test-project",
        )

        assert metadata.friendly_name == "Test table for a non-incremental query"
        assert metadata.description == "Test table for a non-incremental query"
        assert metadata.review_bugs() == ["1999999", "12121212"]

    def test_of_non_existing_table(self):
        with pytest.raises(FileNotFoundError):
            Metadata.of_table(
                "test",
                "no_metadata",
                "v1",
                TEST_DIR / "data" / "test_sql" / "moz-fx-data-test-project",
            )

    def test_is_metadata_file(self):
        assert Metadata.is_metadata_file("foo/bar/invalid.json") is False
        assert Metadata.is_metadata_file("foo/bar/invalid.yaml") is False
        assert Metadata.is_metadata_file("metadata.yaml")
        assert Metadata.is_metadata_file("some/path/to/metadata.yaml")

    def test_set_bigquery_partitioning(self):
        metadata = Metadata.of_table(
            "test",
            "non_incremental_query",
            "v1",
            TEST_DIR / "data" / "test_sql" / "moz-fx-data-test-project",
        )
        metadata.set_bigquery_partitioning(
            field="submission_date",
            partition_type=PartitionType.DAY,
            required=True,
            expiration_days=2,
        )

        assert metadata.bigquery.time_partitioning.field == "submission_date"
        assert metadata.bigquery.time_partitioning.type == PartitionType.DAY
        assert metadata.bigquery.time_partitioning.require_partition_filter
        assert metadata.bigquery.time_partitioning.expiration_days == 2
        assert metadata.bigquery.time_partitioning.expiration_ms == 2 * 86400000

    def test_of_deprecated_metadata(self):
        metadata = Metadata.of_table(
            "test",
            "non_incremental_query",
            "v1",
            TEST_DIR / "data" / "test_sql" / "moz-fx-data-test-project",
        )

        assert metadata.deprecated

    def test_of_dataset_metadata(self):
        metadata = DatasetMetadata.from_file(
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "dataset_metadata.yaml",
        )

        assert metadata.default_table_workgroup_access[0]["members"] == [
            "test_default_member"
        ]
        assert (
            metadata.default_table_workgroup_access[0]["role"]
            == "roles/bigquery.dataViewer"
        )



"""Get the string id in various formats for BigQuery resources."""

import re

FULL_JOB_ID_RE = re.compile(r"(?P<project>[^:]+):(?P<location>[^.]+).(?P<job_id>.+)")


def full_job_id(job):
    """Get the bq cli format fully qualified id for a job."""
    return f"{job.project}:{job.location}.{job.job_id}"


def sql_table_id(table):
    """Get the standard sql format fully qualified id for a table."""
    return f"{table.project}.{table.dataset_id}.{table.table_id}"


def qualified_table_id(table):
    """Get the partially qualified id for a table."""
    return f"{table.dataset_id}.{table.table_id}"



from multiprocessing.pool import ThreadPool
from unittest import mock

import pytest
from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from google.cloud.bigquery import DatasetReference

from bigquery_etl.shredder.config import (
    CLIENT_ID,
    DELETE_TARGETS,
    GLEAN_CLIENT_ID,
    DeleteSource,
    DeleteTarget,
    _list_tables,
    find_glean_targets,
    get_glean_channel_to_app_name_mapping,
)

GLEAN_APP_LISTING = [
    {
        "app_channel": "release",
        "app_id": "org.mozilla.firefox",
        "app_name": "fenix",
        "bq_dataset_family": "org_mozilla_firefox",
    },
    {
        "app_channel": "beta",
        "app_id": "org.mozilla.firefox_beta",
        "app_name": "fenix",
        "bq_dataset_family": "org_mozilla_firefox_beta",
    },
    {
        "app_channel": "nightly",
        "app_id": "org.mozilla.fenix.nightly",
        "app_name": "fenix",
        "bq_dataset_family": "org_mozilla_fenix_nightly",
    },
    {
        "app_channel": "release",
        "app_id": "org.mozilla.focus",
        "app_name": "focus_android",
        "bq_dataset_family": "org_mozilla_focus",
    },
    {
        "app_channel": "beta",
        "app_id": "org.mozilla.focus.beta",
        "app_name": "focus_android",
        "bq_dataset_family": "org_mozilla_focus_beta",
    },
    {
        "app_channel": "beta",
        "app_id": "org.mozilla.focus.beta",
        "app_name": "focus_android",
        "bq_dataset_family": "org_mozilla_focus_beta",
    },
    {
        "app_id": "firefox.desktop",
        "app_name": "firefox_desktop",
        "bq_dataset_family": "firefox_desktop",
    },
]


class FakeClient:
    def list_datasets(self, project):
        return [
            bigquery.DatasetReference(project, f"{app}_{suffix}")
            for app in [
                "org_mozilla_firefox",
                "org_mozilla_firefox_beta",
                "org_mozilla_focus",
                "org_mozilla_focus_beta",
            ]
            for suffix in ["stable", "derived"]
        ]

    def list_tables(self, dataset_ref):
        labels = {}
        if dataset_ref.dataset_id.endswith("stable"):
            table_ids = ["metrics_v1", "deletion_request_v1", "migration_v1"]
            labels["schema_id"] = "glean_ping_1"
        elif dataset_ref.dataset_id in {
            "org_mozilla_focus_derived",
            "org_mozilla_focus_beta_derived",
        }:
            table_ids = [
                "additional_deletion_requests_v1",  # should be ignored
                "clients_daily_v1",
                "dau_v1",  # aggregated, no client_id
            ]
        elif dataset_ref.dataset_id.endswith("derived"):
            table_ids = ["clients_daily_v1"]
        else:
            raise Exception(f"unexpected dataset: {dataset_ref}")
        return [
            bigquery.table.TableListItem(
                {
                    "tableReference": bigquery.TableReference(
                        dataset_ref, table_id
                    ).to_api_repr(),
                    "labels": labels,
                }
            )
            for table_id in table_ids
        ]

    def get_table(self, table_ref):
        table = bigquery.Table(table_ref)
        table._properties[table._PROPERTY_TO_API_FIELD["type"]] = "TABLE"
        if table.dataset_id.endswith("stable"):
            table.schema = [
                bigquery.SchemaField(
                    "client_info",
                    "RECORD",
                    "NULLABLE",
                    [bigquery.SchemaField("client_id", "STRING")],
                )
            ]
        elif table.table_id in {
            "additional_deletion_requests_v1",
            "clients_daily_v1",
        } or table.dataset_id in {
            "fenix",
            "focus_android",
        }:
            table.schema = [bigquery.SchemaField("client_id", "STRING")]
        else:
            table.schema = [bigquery.SchemaField("document_id", "STRING")]
        return table


@mock.patch("bigquery_etl.shredder.config.requests")
def test_glean_targets(mock_requests):
    mock_response = mock.Mock()
    mock_response.json.return_value = GLEAN_APP_LISTING
    mock_requests.get.return_value = mock_response

    with ThreadPool(1) as pool:
        targets = find_glean_targets(pool, FakeClient())

    # convert tuples to sets because additional_deletion_requests are in
    # non-deterministic order due to multiprocessing
    # order doesn't matter in real execution
    for source, target in targets.items():
        targets[source] = set(target)

    assert targets == {
        **{
            target: {  # firefox release tables
                DeleteSource(
                    table="org_mozilla_firefox_stable.deletion_request_v1",
                    field="client_info.client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
            }
            for target in [
                DeleteTarget(
                    table="org_mozilla_firefox_stable.metrics_v1",
                    field=("client_info.client_id",),
                    project="moz-fx-data-shared-prod",
                ),
                DeleteTarget(
                    table="org_mozilla_firefox_derived.clients_daily_v1",
                    field=("client_id",),
                    project="moz-fx-data-shared-prod",
                ),
            ]
        },
        **{
            target: {  # firefox beta tables
                DeleteSource(
                    table="org_mozilla_firefox_beta_stable.deletion_request_v1",
                    field="client_info.client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
            }
            for target in [
                DeleteTarget(
                    table="org_mozilla_firefox_beta_stable.metrics_v1",
                    field=("client_info.client_id",),
                    project="moz-fx-data-shared-prod",
                ),
                DeleteTarget(
                    table="org_mozilla_firefox_beta_derived.clients_daily_v1",
                    field=("client_id",),
                    project="moz-fx-data-shared-prod",
                ),
            ]
        },
        **{
            target: {  # firefox combined-channel tables
                DeleteSource(
                    table="fenix.deletion_request",
                    field="client_info.client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
            }
            for target in [
                DeleteTarget(
                    table="fenix_derived.clients_daily_v1",
                    field=("client_id",),
                    project="moz-fx-data-shared-prod",
                ),
            ]
        },
        **{
            target: {  # focus release tables
                DeleteSource(
                    table="org_mozilla_focus_stable.deletion_request_v1",
                    field="client_info.client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
                DeleteSource(
                    table="org_mozilla_focus_derived.additional_deletion_requests_v1",
                    field="client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
            }
            for target in [
                DeleteTarget(
                    table="org_mozilla_focus_stable.metrics_v1",
                    field=("client_info.client_id",) * 2,
                    project="moz-fx-data-shared-prod",
                ),
                DeleteTarget(
                    table="org_mozilla_focus_derived.clients_daily_v1",
                    field=("client_id",) * 2,
                    project="moz-fx-data-shared-prod",
                ),
            ]
        },
        **{
            target: {  # focus beta tables
                DeleteSource(
                    table="org_mozilla_focus_beta_stable.deletion_request_v1",
                    field="client_info.client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
                DeleteSource(
                    table="org_mozilla_focus_beta_derived.additional_deletion_requests_v1",
                    field="client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
            }
            for target in [
                DeleteTarget(
                    table="org_mozilla_focus_beta_stable.metrics_v1",
                    field=("client_info.client_id",) * 2,
                    project="moz-fx-data-shared-prod",
                ),
                DeleteTarget(
                    table="org_mozilla_focus_beta_derived.clients_daily_v1",
                    field=("client_id",) * 2,
                    project="moz-fx-data-shared-prod",
                ),
            ]
        },
        **{
            target: {  # focus combined-channel tables
                DeleteSource(
                    table="focus_android.deletion_request",
                    field="client_info.client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
                DeleteSource(
                    table="org_mozilla_focus_beta_derived.additional_deletion_requests_v1",
                    field="client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
                DeleteSource(
                    table="org_mozilla_focus_derived.additional_deletion_requests_v1",
                    field="client_id",
                    project="moz-fx-data-shared-prod",
                    conditions=(),
                ),
            }
            for target in [
                DeleteTarget(
                    table="focus_android_derived.clients_daily_v1",
                    field=("client_id",) * 3,
                    project="moz-fx-data-shared-prod",
                ),
            ]
        },
    }


@mock.patch("bigquery_etl.shredder.config.requests")
def test_glean_targets_override(mock_requests):
    """Targets in GLEAN_DERIVED_OVERRIDES should override the target in find_glean_targets."""

    class TargetOverrideClient:
        def list_datasets(self, project):
            return [
                bigquery.DatasetReference(project, name)
                for name in [
                    "firefox_desktop_stable",
                    "firefox_desktop_derived",
                ]
            ]

        def list_tables(self, dataset_ref):
            labels = {}
            if dataset_ref.dataset_id == "firefox_desktop_stable":
                table_ids = ["metrics_v1", "deletion_request_v1"]
                labels["schema_id"] = "glean_ping_1"
            elif dataset_ref.dataset_id == "firefox_desktop_derived":
                table_ids = [
                    "adclick_history_v1",  # should use value from override
                    "other_table_v1",
                    "pageload_1pct_v1",  # should be ignored
                ]
            else:
                raise Exception(f"unexpected dataset: {dataset_ref}")
            return [
                bigquery.table.TableListItem(
                    {
                        "tableReference": bigquery.TableReference(
                            dataset_ref, table_id
                        ).to_api_repr(),
                        "labels": labels,
                    }
                )
                for table_id in table_ids
            ]

        def get_table(self, table_ref):
            table = bigquery.Table(table_ref)
            table._properties[table._PROPERTY_TO_API_FIELD["type"]] = "TABLE"
            if table.dataset_id.endswith("stable"):
                table.schema = [
                    bigquery.SchemaField(
                        "client_info",
                        "RECORD",
                        "NULLABLE",
                        [bigquery.SchemaField("client_id", "STRING")],
                    )
                ]
            else:
                table.schema = [bigquery.SchemaField("client_id", "STRING")]
            return table

    mock_response = mock.Mock()
    mock_response.json.return_value = GLEAN_APP_LISTING
    mock_requests.get.return_value = mock_response

    with ThreadPool(1) as pool:
        targets = find_glean_targets(pool, TargetOverrideClient())

    # convert tuples to sets because additional_deletion_requests are in
    # non-deterministic order due to multiprocessing
    # order doesn't matter in real execution
    for source, target in targets.items():
        targets[source] = set(target) if isinstance(target, tuple) else target

    desktop_deletions = DeleteSource(
        table="firefox_desktop_stable.deletion_request_v1",
        field=GLEAN_CLIENT_ID,
    )

    assert targets == {
        DeleteTarget(
            table="firefox_desktop_derived.other_table_v1", field=(CLIENT_ID,)
        ): {desktop_deletions},
        DeleteTarget(
            table="firefox_desktop_stable.metrics_v1", field=(GLEAN_CLIENT_ID,)
        ): {desktop_deletions},
        # adclick_history_v1 should not be here
    }


@mock.patch("bigquery_etl.shredder.config.requests")
def test_glean_channel_app_mapping(mock_requests):
    mock_response = mock.Mock()
    mock_response.json.return_value = GLEAN_APP_LISTING
    mock_requests.get.return_value = mock_response

    actual = get_glean_channel_to_app_name_mapping()

    expected = {
        "org_mozilla_firefox": "fenix",
        "org_mozilla_firefox_beta": "fenix",
        "org_mozilla_fenix_nightly": "fenix",
        "org_mozilla_focus": "focus_android",
        "org_mozilla_focus_beta": "focus_android",
        "firefox_desktop": "firefox_desktop",
    }

    assert actual == expected


def test_list_tables_wrapper_empty():
    """List tables wrapper should return an empty list when dataset doesn't exist."""

    class EmptyFakeClient(FakeClient):
        def list_tables(self, dataset_ref):
            raise NotFound("not found")

    tables = _list_tables(DatasetReference("project", "dataset"), EmptyFakeClient())

    assert tables == []


def test_delete_target_fields_match_sources():
    """The number of fields in the delete targets should match the number of sources."""
    for target, sources in DELETE_TARGETS.items():
        field_count = 1 if isinstance(target.field, str) else len(target.field)
        source_count = 1 if isinstance(sources, DeleteSource) else len(sources)
        assert field_count == source_count, (
            f"Invalid delete target for {target.table}: number of fields in target "
            f"(found {field_count}) must match number of sources (found {source_count})"
        )


def test_delete_source_invalid():
    """DeleteSource constructor should fail when the given table is invalid."""
    DeleteSource(
        table="dataset.deletion_request_v1",
        field="client_id",
        project="project",
    )

    with pytest.raises(ValueError):
        DeleteSource(
            table="deletion_request_v1",
            field="client_id",
            project="project",
        )

    with pytest.raises(ValueError):
        DeleteSource(
            table="project.dataset.deletion_request_v1",
            field="client_id",
            project="project",
        )


def test_delete_target_invalid():
    """DeleteTarget constructor should fail when the given table is invalid."""
    DeleteTarget(
        table="dataset.deletion_request_v1",
        field="client_id",
        project="project",
    )

    with pytest.raises(ValueError):
        DeleteTarget(
            table="deletion_request_v1",
            field="client_id",
            project="project",
        )

    with pytest.raises(ValueError):
        DeleteTarget(
            table="project.dataset.deletion_request_v1",
            field="client_id",
            project="project",
        )



"""bigquery-etl CLI backfill command."""

import json
import logging
import subprocess
import sys
import tempfile
from datetime import date, datetime, timedelta
from pathlib import Path

import rich_click as click
import yaml
from google.cloud import bigquery
from google.cloud.exceptions import Conflict, NotFound

from ..backfill.date_range import BackfillDateRange, get_backfill_partition
from ..backfill.parse import (
    BACKFILL_FILE,
    DEFAULT_BILLING_PROJECT,
    DEFAULT_REASON,
    DEFAULT_WATCHER,
    Backfill,
    BackfillStatus,
)
from ..backfill.shredder_mitigation import (
    SHREDDER_MITIGATION_CHECKS_NAME,
    SHREDDER_MITIGATION_QUERY_NAME,
    generate_query_with_shredder_mitigation,
)
from ..backfill.utils import (
    get_backfill_backup_table_name,
    get_backfill_file_from_qualified_table_name,
    get_backfill_staging_qualified_table_name,
    get_entries_from_qualified_table_name,
    get_qualified_table_name_to_entries_map_by_project,
    get_scheduled_backfills,
    qualified_table_name_matching,
    validate_depends_on_past,
    validate_metadata_workgroups,
)
from ..backfill.validate import (
    validate_duplicate_entry_with_initiate_status,
    validate_file,
)
from ..cli.query import backfill as query_backfill
from ..cli.utils import (
    billing_project_option,
    is_authenticated,
    project_id_option,
    sql_dir_option,
)
from ..config import ConfigLoader
from ..deploy import FailedDeployException, SkippedDeployException, deploy_table
from ..metadata.parse_metadata import METADATA_FILE, Metadata
from ..metadata.validate_metadata import SHREDDER_MITIGATION_LABEL
from ..schema import SCHEMA_FILE, Schema

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


@click.group(help="Commands for managing backfills.")
@click.pass_context
def backfill(ctx):
    """Create the CLI group for the backfill command."""
    # create temporary directory generated content is written to
    # the directory will be deleted automatically after the command exits
    ctx.ensure_object(dict)
    ctx.obj["TMP_DIR"] = ctx.with_resource(tempfile.TemporaryDirectory())


@backfill.command(
    help="""Create a new backfill entry in the backfill.yaml file.  Create
    a backfill.yaml file if it does not already exist.

    Examples:

    \b
    ./bqetl backfill create moz-fx-data-shared-prod.telemetry_derived.deviations_v1 \\
      --start_date=2021-03-01 \\
      --end_date=2021-03-31 \\
      --exclude=2021-03-03 \\
    """,
)
@click.argument("qualified_table_name")
@sql_dir_option
@click.option(
    "--start_date",
    "--start-date",
    "-s",
    help="First date to be backfilled. Date format: yyyy-mm-dd",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    required=True,
)
@click.option(
    "--end_date",
    "--end-date",
    "-e",
    help="Last date to be backfilled. Date format: yyyy-mm-dd",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=datetime.today(),
)
@click.option(
    "--exclude",
    "-x",
    multiple=True,
    help="Dates excluded from backfill. Date format: yyyy-mm-dd",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=None,
)
@click.option(
    "--watcher",
    "-w",
    help="Watcher of the backfill (email address)",
    default=DEFAULT_WATCHER,
)
@click.option(
    "--custom_query_path",
    "--custom-query-path",
    help="Path of the custom query to run the backfill. Optional.",
)
@click.option(
    "--shredder_mitigation/--no_shredder_mitigation",
    help="Wether to run a backfill using an auto-generated query that mitigates shredder effect.",
)
@click.option(
    "--override-retention-range-limit",
    "--override_retention_range_limit",
    required=False,
    type=bool,
    is_flag=True,
    help="True to allow running a backfill outside the retention policy limit.",
    default=False,
)
# If not specified, the billing project will be set to the default billing project when the backfill is initiated.
@billing_project_option()
@click.pass_context
def create(
    ctx,
    qualified_table_name,
    sql_dir,
    start_date,
    end_date,
    exclude,
    watcher,
    custom_query_path,
    shredder_mitigation,
    override_retention_range_limit,
    billing_project,
):
    """CLI command for creating a new backfill entry in backfill.yaml file.

    A backfill.yaml file will be created if it does not already exist.
    """
    if not validate_depends_on_past(sql_dir, qualified_table_name):
        click.echo("Tables that depend on past are currently not supported.")
        sys.exit(1)

    if not validate_metadata_workgroups(sql_dir, qualified_table_name):
        click.echo("Only mozilla-confidential workgroups are supported.")
        sys.exit(1)

    existing_backfills = get_entries_from_qualified_table_name(
        sql_dir, qualified_table_name
    )

    new_entry = Backfill(
        entry_date=date.today(),
        start_date=start_date.date(),
        end_date=end_date.date(),
        excluded_dates=[e.date() for e in list(exclude)],
        reason=DEFAULT_REASON,
        watchers=[watcher],
        status=BackfillStatus.INITIATE,
        custom_query_path=custom_query_path,
        shredder_mitigation=shredder_mitigation,
        override_retention_limit=override_retention_range_limit,
        billing_project=billing_project,
    )

    validate_duplicate_entry_with_initiate_status(new_entry, existing_backfills)

    existing_backfills.insert(0, new_entry)

    backfill_file = get_backfill_file_from_qualified_table_name(
        sql_dir, qualified_table_name
    )

    backfill_file.write_text(
        "\n".join(
            backfill.to_yaml() for backfill in sorted(existing_backfills, reverse=True)
        )
    )

    click.echo(f"Created backfill entry in {backfill_file}.")


@backfill.command(
    help="""Validate backfill.yaml file format and content.

    Examples:

    ./bqetl backfill validate moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6

    \b
    # validate all backfill.yaml files if table is not specified
    Use the `--project_id` option to change the project to be validated;
    default is `moz-fx-data-shared-prod`.

    Examples:

    ./bqetl backfill validate
    """
)
@click.argument("qualified_table_name", required=False)
@sql_dir_option
@project_id_option()
@click.pass_context
def validate(
    ctx,
    qualified_table_name,
    sql_dir,
    project_id,
):
    """Validate backfill.yaml files."""
    if qualified_table_name:
        backfills_dict = {
            qualified_table_name: get_entries_from_qualified_table_name(
                sql_dir, qualified_table_name
            )
        }
    else:
        backfills_dict = get_qualified_table_name_to_entries_map_by_project(
            sql_dir, project_id
        )

    for table_name in backfills_dict:

        if not validate_depends_on_past(sql_dir, table_name):
            click.echo(
                f"Tables that depend on past are currently not supported:  {table_name}"
            )
            sys.exit(1)

        if not validate_metadata_workgroups(sql_dir, table_name):
            click.echo(
                f"Only mozilla-confidential workgroups are supported.  {table_name} contain workgroup access that is not supported"
            )
            sys.exit(1)

        try:
            backfill_file = get_backfill_file_from_qualified_table_name(
                sql_dir, table_name
            )
            validate_file(backfill_file)
        except (yaml.YAMLError, ValueError) as e:
            click.echo(
                f"Backfill.yaml file for {table_name} contains the following error:\n {e}"
            )
            sys.exit(1)

        click.echo(f"{BACKFILL_FILE} has been validated for {table_name}.")

    if backfills_dict:
        click.echo(
            f"All {BACKFILL_FILE} files have been validated for project {project_id}."
        )


@backfill.command(
    help="""Get backfill(s) information from all or specific table(s).

    Examples:

    # Get info for specific table.
    ./bqetl backfill info moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6

    \b
    # Get info for all tables.
    ./bqetl backfill info

    \b
    # Get info from all tables with specific status.
    ./bqetl backfill info --status=Initiate
    """,
)
@click.argument("qualified_table_name", required=False)
@sql_dir_option
@project_id_option(
    ConfigLoader.get("default", "project", fallback="moz-fx-data-shared-prod")
)
@click.option(
    "--status",
    type=click.Choice([s.value for s in BackfillStatus]),
    help="Filter backfills with this status.",
)
@click.pass_context
def info(ctx, qualified_table_name, sql_dir, project_id, status):
    """Return backfill(s) information from all or specific table(s)."""
    if qualified_table_name:
        backfills_dict = {
            qualified_table_name: get_entries_from_qualified_table_name(
                sql_dir, qualified_table_name, status
            )
        }
    else:
        backfills_dict = get_qualified_table_name_to_entries_map_by_project(
            sql_dir, project_id, status
        )

    total_backfills_count = 0

    for qualified_table_name, entries in backfills_dict.items():
        entries_count = len(entries)

        total_backfills_count += entries_count

        project, dataset, table = qualified_table_name_matching(qualified_table_name)
        status_str = f" with {status} status" if status is not None else ""
        click.echo(
            f"""{project}.{dataset}.{table} has {entries_count} backfill(s){status_str}:"""
        )

        for entry in entries:
            click.echo(str(entry))

    click.echo(f"\nThere are a total of {total_backfills_count} backfill(s).")


@backfill.command(
    help="""Get information on backfill(s) that require processing.

    Examples:

    # Get info for specific table.
    ./bqetl backfill scheduled moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6

    \b
    # Get info for all tables.
    ./bqetl backfill scheduled
    """,
)
@click.argument("qualified_table_name", required=False)
@sql_dir_option
@project_id_option(
    ConfigLoader.get("default", "project", fallback="moz-fx-data-shared-prod")
)
@click.option(
    "--status",
    type=click.Choice([s.value for s in BackfillStatus]),
    default=BackfillStatus.INITIATE.value,
    help="Whether to get backfills to process or to complete.",
)
@click.option("--json_path", type=click.Path())
@click.pass_context
def scheduled(ctx, qualified_table_name, sql_dir, project_id, status, json_path=None):
    """Return list of backfill(s) that require processing."""
    backfills = get_scheduled_backfills(
        sql_dir, project_id, qualified_table_name, status=status
    )

    for qualified_table_name, entry in backfills.items():
        click.echo(f"Backfill scheduled for {qualified_table_name}:\n{entry}")

    click.echo(f"{len(backfills)} backfill(s) require processing.")

    if json_path is not None:
        formatted_backfills = [
            {
                "qualified_table_name": qualified_table_name,
                "entry_date": entry.entry_date.strftime("%Y-%m-%d"),
                "watchers": entry.watchers,
            }
            for qualified_table_name, entry in backfills.items()
        ]

        Path(json_path).write_text(json.dumps(formatted_backfills))


@backfill.command(
    help="""Process entry in backfill.yaml with Initiate status that has not yet been processed.

    Examples:

    \b

    # Initiate backfill entry for specific table
    ./bqetl backfill initiate moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6

    Use the `--project_id` option to change the project;
    default project_id is `moz-fx-data-shared-prod`.
    """
)
@click.argument("qualified_table_name")
@click.option(
    "--parallelism",
    default=16,
    type=int,
    help="Maximum number of queries to execute concurrently",
)
@sql_dir_option
@project_id_option(
    ConfigLoader.get("default", "project", fallback="moz-fx-data-shared-prod")
)
@click.pass_context
def initiate(
    ctx,
    qualified_table_name,
    parallelism,
    sql_dir,
    project_id,
):
    """Process backfill entry with initiate status in backfill.yaml file(s)."""
    click.echo("Backfill processing (initiate) started....")

    backfills_to_process_dict = get_scheduled_backfills(
        sql_dir, project_id, qualified_table_name, status=BackfillStatus.INITIATE.value
    )

    if not backfills_to_process_dict:
        click.echo(f"No backfill processed for {qualified_table_name}")
        return

    entry_to_initiate = backfills_to_process_dict[qualified_table_name]

    backfill_staging_qualified_table_name = get_backfill_staging_qualified_table_name(
        qualified_table_name, entry_to_initiate.entry_date
    )

    project, dataset, table = qualified_table_name_matching(qualified_table_name)
    query_path = Path(sql_dir) / project / dataset / table / "query.sql"

    # create schema before deploying staging table if it does not exist
    schema_path = query_path.parent / SCHEMA_FILE

    if not schema_path.exists():
        # if schema doesn't exist, a schema file is created to allow backfill staging table deployment
        Schema.from_query_file(
            query_file=query_path,
            respect_skip=False,
            sql_dir=sql_dir,
        ).to_yaml_file(schema_path)
        click.echo(f"Schema file created for {qualified_table_name}: {schema_path}")

    try:
        deploy_table(
            artifact_file=query_path,
            destination_table=backfill_staging_qualified_table_name,
            respect_dryrun_skip=False,
        )
    except (SkippedDeployException, FailedDeployException) as e:
        raise RuntimeError(
            f"Backfill initiate failed to deploy {query_path} to {backfill_staging_qualified_table_name}."
        ) from e

    billing_project = DEFAULT_BILLING_PROJECT

    # override with billing project from backfill entry
    if entry_to_initiate.billing_project is not None:
        billing_project = entry_to_initiate.billing_project
    elif not billing_project.startswith("moz-fx-data-backfill-"):
        raise ValueError(
            f"Invalid billing project: {billing_project}.  Please use one of the projects assigned to backfills."
        )
        sys.exit(1)

    click.echo(
        f"\nInitiating backfill for {qualified_table_name} with entry date {entry_to_initiate.entry_date} via dry run:"
    )

    _initiate_backfill(
        ctx,
        qualified_table_name,
        backfill_staging_qualified_table_name,
        entry_to_initiate,
        parallelism,
        dry_run=True,
        billing_project=billing_project,
    )

    click.echo(
        f"\nInitiating backfill for {qualified_table_name} with entry date {entry_to_initiate.entry_date}:"
    )
    _initiate_backfill(
        ctx,
        qualified_table_name,
        backfill_staging_qualified_table_name,
        entry_to_initiate,
        parallelism,
        billing_project=billing_project,
    )

    click.echo(
        f"Processed backfill for {qualified_table_name} with entry date {entry_to_initiate.entry_date}"
    )


def _initiate_backfill(
    ctx,
    qualified_table_name: str,
    backfill_staging_qualified_table_name: str,
    entry: Backfill,
    parallelism: int = 16,
    dry_run: bool = False,
    billing_project=DEFAULT_BILLING_PROJECT,
):
    if not is_authenticated():
        click.echo(
            "Authentication to GCP required. Run `gcloud auth login  --update-adc` "
            "and check that the project is set correctly."
        )
        sys.exit(1)

    project, dataset, table = qualified_table_name_matching(qualified_table_name)

    logging_str = f"""Initiating backfill for {qualified_table_name} (destination: {backfill_staging_qualified_table_name}).
                    Query will be executed in {billing_project}."""

    if dry_run:
        logging_str += "  This is a dry run."

    log.info(logging_str)

    custom_query_path = None
    checks = None
    custom_checks_name = None

    metadata = Metadata.from_file(
        Path("sql") / project / dataset / table / METADATA_FILE
    )

    # Stop if the metadata contains shredder mitigation label and the backfill doesn't.
    if (
        SHREDDER_MITIGATION_LABEL in metadata.labels
        and entry.shredder_mitigation is False
    ):
        click.echo(
            click.style(
                f"This backfill cannot continue.\nManaged backfills for tables with metadata label"
                f" {SHREDDER_MITIGATION_LABEL} require using --shredder_mitigation.",
                fg="yellow",
            )
        )
        sys.exit(1)

    if entry.shredder_mitigation is True:
        click.echo(
            click.style(
                f"Generating query with shredder mitigation for {dataset}.{table}...",
                fg="blue",
            )
        )
        query_path, _ = generate_query_with_shredder_mitigation(
            client=bigquery.Client(project=project),
            project_id=project,
            dataset=dataset,
            destination_table=table,
            staging_table_name=backfill_staging_qualified_table_name,
            backfill_date=entry.start_date.isoformat(),
        )
        custom_query_path = Path(query_path) / f"{SHREDDER_MITIGATION_QUERY_NAME}.sql"
        checks = True
        custom_checks_name = f"{SHREDDER_MITIGATION_CHECKS_NAME}.sql"
        click.echo(
            click.style(
                f"Starting backfill with custom query: '{custom_query_path}'.",
                fg="blue",
            )
        )
    elif entry.custom_query_path:
        custom_query_path = Path(entry.custom_query_path)

    override_retention_limit = entry.override_retention_limit

    # Backfill table
    # in the long-run we should remove the query backfill command and require a backfill entry for all backfills
    try:
        ctx.invoke(
            query_backfill,
            name=f"{dataset}.{table}",
            project_id=project,
            start_date=datetime.fromisoformat(entry.start_date.isoformat()),
            end_date=datetime.fromisoformat(entry.end_date.isoformat()),
            exclude=[e.strftime("%Y-%m-%d") for e in entry.excluded_dates],
            destination_table=backfill_staging_qualified_table_name,
            parallelism=parallelism,
            dry_run=dry_run,
            **(
                {
                    k: param
                    for k, param in [
                        ("custom_query_path", custom_query_path),
                        ("checks", checks),
                        ("checks_file_name", custom_checks_name),
                    ]
                    if param is not None
                }
            ),
            billing_project=billing_project,
            override_retention_range_limit=override_retention_limit,
        )
    except subprocess.CalledProcessError as e:
        raise ValueError(
            f"Backfill initiate resulted in error for {qualified_table_name}"
        ) from e


@backfill.command(
    help="""Complete entry in backfill.yaml with Complete status that has not yet been processed..

    Examples:

    \b

    # Complete backfill entry for specific table
    ./bqetl backfill complete moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6

    Use the `--project_id` option to change the project;
    default project_id is `moz-fx-data-shared-prod`.
    """
)
@click.argument("qualified_table_name")
@sql_dir_option
@project_id_option("moz-fx-data-shared-prod")
@click.pass_context
def complete(ctx, qualified_table_name, sql_dir, project_id):
    """Process backfill entry with complete status in backfill.yaml file(s)."""
    if not is_authenticated():
        click.echo(
            "Authentication to GCP required. Run `gcloud auth login  --update-adc` "
            "and check that the project is set correctly."
        )
        sys.exit(1)
    client = bigquery.Client(project=project_id)

    click.echo("Backfill processing (complete) started....")

    backfills_to_process_dict = get_scheduled_backfills(
        sql_dir, project_id, qualified_table_name, status=BackfillStatus.COMPLETE.value
    )

    if not backfills_to_process_dict:
        click.echo(f"No backfill processed for {qualified_table_name}")
        return

    entry_to_complete = backfills_to_process_dict[qualified_table_name]

    click.echo(
        f"Completing backfill for {qualified_table_name} with entry date {entry_to_complete.entry_date}:"
    )

    backfill_staging_qualified_table_name = get_backfill_staging_qualified_table_name(
        qualified_table_name, entry_to_complete.entry_date
    )

    # clone production table
    cloned_table_full_name = get_backfill_backup_table_name(
        qualified_table_name, entry_to_complete.entry_date
    )
    _copy_table(qualified_table_name, cloned_table_full_name, client, clone=True)

    project, dataset, table = qualified_table_name_matching(qualified_table_name)
    table_metadata = Metadata.from_file(
        Path(sql_dir) / project / dataset / table / METADATA_FILE
    )

    _copy_backfill_staging_to_prod(
        backfill_staging_qualified_table_name,
        qualified_table_name,
        client,
        entry_to_complete,
        table_metadata,
    )

    # delete backfill staging table
    client.delete_table(backfill_staging_qualified_table_name)
    click.echo(
        f"Backfill staging table deleted: {backfill_staging_qualified_table_name}"
    )

    click.echo(
        f"Processed backfill for {qualified_table_name} with entry date {entry_to_complete.entry_date}"
    )


def _copy_backfill_staging_to_prod(
    backfill_staging_table: str,
    qualified_table_name: str,
    client: bigquery.Client,
    entry: Backfill,
    table_metadata: Metadata,
):
    """Copy backfill staging table to prod based on table metadata and backfill config.

    If table is
       un-partitioned: copy the entire staging table to production.
       partitioned: determine and copy each partition from staging to production.
    """
    partitioning_type = None
    if table_metadata.bigquery and table_metadata.bigquery.time_partitioning:
        partitioning_type = table_metadata.bigquery.time_partitioning.type

    if partitioning_type is None:
        _copy_table(backfill_staging_table, qualified_table_name, client)
    else:
        backfill_date_range = BackfillDateRange(
            entry.start_date,
            entry.end_date,
            excludes=entry.excluded_dates,
            range_type=partitioning_type,
        )
        # If date_partition_parameter isn't set it's assumed to be submission_date:
        # https://github.com/mozilla/telemetry-airflow/blob/dbc2782fa23a34ae8268e7788f9621089ac71def/utils/gcp.py#L194C48-L194C48
        partition_param, offset = "submission_date", 0
        if table_metadata.scheduling:
            partition_param = table_metadata.scheduling.get(
                "date_partition_parameter", partition_param
            )
            offset = table_metadata.scheduling.get("date_partition_offset", offset)

        for backfill_date in backfill_date_range:
            if (
                partition := get_backfill_partition(
                    backfill_date,
                    partition_param,
                    offset,
                    partitioning_type,
                )
            ) is None:
                raise ValueError(
                    f"Null partition found completing backfill {entry} for {qualified_table_name}."
                )

            production_table = f"{qualified_table_name}${partition}"
            backfill_table = f"{backfill_staging_table}${partition}"
            _copy_table(backfill_table, production_table, client)


def _copy_table(
    source_table: str, destination_table: str, client, clone: bool = False
) -> None:
    """
    Copy and overwrite table from source to destination table.

    If clone is True, clone (previous) production data for backup before swapping stage data into production.
    """
    job_type_str = "copied"

    if clone:
        copy_config = bigquery.CopyJobConfig(
            write_disposition=bigquery.WriteDisposition.WRITE_EMPTY,
            operation_type=bigquery.job.copy_.OperationType.CLONE,
            destination_expiration_time=(datetime.now() + timedelta(days=30)).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            ),
        )
        job_type_str = "cloned"
    else:
        copy_config = bigquery.CopyJobConfig(
            write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
            operation_type=bigquery.job.copy_.OperationType.COPY,
        )

    try:
        client.copy_table(
            source_table,
            destination_table,
            job_config=copy_config,
        ).result()
    except NotFound:
        click.echo(f"Source table not found: {source_table}")
        sys.exit(1)
    except Conflict:
        click.echo(f"Backup table already exists: {destination_table}")
        sys.exit(1)

    click.echo(
        f"Table {source_table} successfully {job_type_str} to {destination_table}"
    )



import csv
import os
from pathlib import Path

from bigquery_etl.util.common import snake_case


def snake_case_test(case_name: str):
    resource_path = Path("tests/resources/casing").resolve()
    test_file = os.path.join(resource_path, case_name)

    with open(test_file) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        for row in csv_reader:
            assert len(row) == 2
            assert snake_case(row[0]) == row[1]


def test_snake_casing():
    # all strings of length 3 drawn from the alphabet "aA7"
    snake_case_test("alphanum_3.csv")

    # all strings of length 4 drawn from the alphabet "aA7_"
    snake_case_test("word_4.csv")

    # all column names from mozilla-pipeline-schemas affected by snake_casing
    # https://github.com/mozilla/jsonschema-transpiler/pull/79#issuecomment-509839572
    # https://gist.github.com/acmiyaguchi/3f526c440b67ebe469bcb6ab2da5123f#file-readme-md
    snake_case_test("mps-diff-integration.csv")



#!/usr/bin/env python3

"""Load Pocket sponsored content metadata from GCS."""
import uuid
from argparse import ArgumentParser
from datetime import datetime, timedelta

from google.cloud import bigquery

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--date", type=lambda x: datetime.strptime(x, "%Y-%m-%d").date(), required=True
)
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--source-bucket", default="moz-fx-data-prod-external-pocket-data")
parser.add_argument("--source-prefix", default="spoc_tile_ids")
parser.add_argument("--destination_dataset", default="pocket_derived")
parser.add_argument("--destination_table", default="spoc_tile_ids_history_v1")


def main():
    """Load data to temp table and populate destination table via query."""
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    # The Pocket-side ETL uses dbt and includes partial days, so need some handling
    # for that; see https://bugzilla.mozilla.org/show_bug.cgi?id=1695336#c33
    #
    # The Airflow DAG run for 2021-03-01 expects to process complete results up
    # through 2021-03-01, and executes during UTC day 2021-03-02. That run
    # will pick up a Pocket result labeled as 2021-03-02, so we need to add an
    # offset here. Then we make sure to drop partial data for 2021-03-02
    # in the final query.
    publish_date = (args.date + timedelta(days=1)).isoformat()

    # First, we load the data from GCS to a temp table.
    uri = f"gs://{args.source_bucket}/{args.source_prefix}/{publish_date}/*"
    tmp_suffix = uuid.uuid4().hex[0:6]
    tmp_table = f"tmp.pocket_spoc_tile_ids_load_{tmp_suffix}"
    client.load_table_from_uri(
        uri,
        tmp_table,
        job_config=bigquery.LoadJobConfig(
            write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE,
            source_format=bigquery.job.SourceFormat.PARQUET,
        ),
    ).result()

    # Next, we run a query that populates the destination table partition.
    client.query(
        f"""
        SELECT DATE("{args.date.isoformat()}") AS submission_date,
        CAST(tile_id AS INT64) AS tile_id,
        type,
        FROM {tmp_table}
        """,
        job_config=bigquery.QueryJobConfig(
            destination=".".join(
                [
                    args.project,
                    args.destination_dataset,
                    f"{args.destination_table}${args.date.strftime('%Y%m%d')}",
                ]
            ),
            write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE,
        ),
    ).result()

    # Finally, we clean up after ourselves.
    client.delete_table(tmp_table)


if __name__ == "__main__":
    main()



"""bigquery-etl CLI stage commands."""

import re
import shutil
import tempfile
from datetime import datetime
from glob import glob
from pathlib import Path

import rich_click as click
from google.cloud import bigquery
from google.cloud.bigquery.enums import EntityTypes
from google.cloud.exceptions import NotFound

from .. import ConfigLoader
from ..cli.query import deploy as deploy_query_schema
from ..cli.query import update as update_query_schema
from ..cli.routine import publish as publish_routine
from ..cli.utils import paths_matching_name_pattern, sql_dir_option
from ..cli.view import publish as publish_view
from ..dryrun import DryRun, get_id_token
from ..routine.parse_routine import (
    ROUTINE_FILES,
    UDF_FILE,
    RawRoutine,
    accumulate_dependencies,
    read_routine_dir,
)
from ..schema import SCHEMA_FILE, Schema
from ..util.common import render
from ..view import View

VIEW_FILE = "view.sql"
QUERY_FILE = "query.sql"
QUERY_SCRIPT = "query.py"
MATERIALIZED_VIEW = "materialized_view.sql"
ROOT = Path(__file__).parent.parent.parent
TEST_DIR = ROOT / "tests" / "sql"


@click.group(help="Commands for managing stage deploys")
def stage():
    """Create the CLI group for the stage command."""
    pass


@stage.command(
    help="""
        Deploy artifacts to the configured stage project. The order of deployment is:
        UDFs, views, tables.

    Examples:
    ./bqetl stage deploy sql/moz-fx-data-shared-prod/telemetry_derived/
    """
)
@click.argument(
    "paths",
    nargs=-1,
)
@click.option(
    "--project-id",
    "--project_id",
    help="GCP project to deploy artifacts to",
    default="bigquery-etl-integration-test",
)
@sql_dir_option
@click.option(
    "--dataset-suffix",
    "--dataset_suffix",
    help="Suffix appended to the deployed dataset",
)
@click.option(
    "--update-references",
    "--update_references",
    default=True,
    is_flag=True,
    help="Update references to deployed artifacts with project and dataset artifacts have been deployed to.",
)
@click.option(
    "--copy-sql-to-tmp-dir",
    "--copy_sql_to_tmp_dir",
    help="Copy existing SQL from the sql_dir to a temporary directory and apply updates there.",
    default=False,
    is_flag=True,
)
@click.option(
    "--remove-updated-artifacts",
    "--remove_updated_artifacts",
    default=False,
    is_flag=True,
    help="Remove artifacts that have been updated and deployed to stage from prod folder. This ensures that"
    + " tests don't run on outdated or undeployed artifacts (required for CI)",
)
@click.pass_context
def deploy(
    ctx,
    paths,
    project_id,
    sql_dir,
    dataset_suffix,
    update_references,
    copy_sql_to_tmp_dir,
    remove_updated_artifacts,
):
    """Deploy provided artifacts to destination project."""
    if copy_sql_to_tmp_dir:
        # copy SQL to a temporary directory
        tmp_dir = Path(tempfile.mkdtemp())
        tmp_dir.mkdir(parents=True, exist_ok=True)
        new_sql_dir = tmp_dir / Path(sql_dir).name
        shutil.copytree(sql_dir, new_sql_dir, dirs_exist_ok=True)

        # rename paths to tmp_dir
        paths = [path.replace(sql_dir, f"{new_sql_dir}/", 1) for path in paths]

        sql_dir = new_sql_dir

    artifact_files = set()

    # get SQL files for artifacts that are to be deployed
    for path in paths:
        artifact_files.update(
            [
                p
                for p in paths_matching_name_pattern(
                    path, sql_dir, None, files=["*.sql", "*.py"]
                )
                if p.suffix in [".sql", ".py"]
                and p.name != "checks.sql"
                and len(p.suffixes) == 1
            ]
        )

    # any dependencies need to be determined an deployed as well since the stage
    # environment doesn't have access to the prod environment
    artifact_files.update(_udf_dependencies(artifact_files))
    artifact_files.update(_view_dependencies(artifact_files, sql_dir))

    # update references of all deployed artifacts
    # references needs to be set to the stage project and the new dataset identifier
    if update_references:
        _update_references(artifact_files, project_id, dataset_suffix, sql_dir)

    updated_artifact_files = set()
    (Path(sql_dir) / project_id).mkdir(parents=True, exist_ok=True)
    # copy updated files locally to a folder representing the stage env project
    for artifact_file in artifact_files:
        artifact_project = artifact_file.parent.parent.parent.name
        artifact_dataset = artifact_file.parent.parent.name
        artifact_name = artifact_file.parent.name
        artifact_test_path = (
            TEST_DIR / artifact_project / artifact_dataset / artifact_name
        )

        if (
            artifact_dataset == "INFORMATION_SCHEMA"
            or "INFORMATION_SCHEMA" in artifact_name
        ):
            continue

        new_artifact_dataset = (
            f"{artifact_dataset}_{artifact_project.replace('-', '_')}"
        )
        if dataset_suffix:
            new_artifact_dataset = f"{new_artifact_dataset}_{dataset_suffix}"

        if artifact_file.name == MATERIALIZED_VIEW:
            # replace CREATE MATERIALIED VIEW statement
            sql_content = render(
                artifact_file.name,
                template_folder=str(artifact_file.parent),
                format=False,
            )
            sql_content = re.sub(
                "CREATE MATERIALIZED VIEW.*?AS",
                "",
                sql_content,
                flags=re.DOTALL,
            )
            artifact_file.write_text(sql_content)
            # map materialized views to normal queries
            query_path = Path(artifact_file.parent, QUERY_FILE)
            artifact_file.rename(query_path)
            artifact_file = query_path

        new_artifact_path = (
            Path(sql_dir) / project_id / new_artifact_dataset / artifact_name
        )
        new_artifact_path.mkdir(parents=True, exist_ok=True)
        shutil.copytree(artifact_file.parent, new_artifact_path, dirs_exist_ok=True)
        updated_artifact_files.add(new_artifact_path / artifact_file.name)

        # copy tests to the right structure
        if artifact_test_path.exists():
            new_artifact_test_path = (
                TEST_DIR / project_id / new_artifact_dataset / artifact_name
            )
            shutil.copytree(
                artifact_test_path, new_artifact_test_path, dirs_exist_ok=True
            )
            if remove_updated_artifacts:
                shutil.rmtree(artifact_test_path)

    # rename test files
    for test_file_path in map(Path, glob(f"{TEST_DIR}/**/*", recursive=True)):
        test_file_suffix = test_file_path.suffix
        for artifact_file in artifact_files:
            artifact_project = artifact_file.parent.parent.parent.name
            artifact_dataset = artifact_file.parent.parent.name
            artifact_name = artifact_file.parent.name
            if test_file_path.name in (
                f"{artifact_project}.{artifact_dataset}.{artifact_name}{test_file_suffix}",
                f"{artifact_project}.{artifact_dataset}.{artifact_name}.schema{test_file_suffix}",
            ) or (
                test_file_path.name
                in (
                    f"{artifact_dataset}.{artifact_name}{test_file_suffix}",
                    f"{artifact_dataset}.{artifact_name}.schema{test_file_suffix}",
                )
                and artifact_project in test_file_path.parent.parts
            ):
                new_artifact_dataset = (
                    f"{artifact_dataset}_{artifact_project.replace('-', '_')}"
                )
                if dataset_suffix:
                    new_artifact_dataset = f"{new_artifact_dataset}_{dataset_suffix}"

                new_test_file_name = (
                    f"{project_id}.{new_artifact_dataset}.{artifact_name}"
                )
                if test_file_path.name.endswith(f".schema{test_file_suffix}"):
                    new_test_file_name += ".schema"
                new_test_file_name += test_file_suffix

                new_test_file_path = test_file_path.parent / new_test_file_name
                if not new_test_file_path.exists():
                    test_file_path.rename(new_test_file_path)
                break

    # remove artifacts from the "prod" folders
    if remove_updated_artifacts:
        for artifact_file in artifact_files:
            if artifact_file.parent.exists():
                shutil.rmtree(artifact_file.parent)

    # deploy to stage
    _deploy_artifacts(ctx, updated_artifact_files, project_id, dataset_suffix, sql_dir)


def _udf_dependencies(artifact_files):
    """Determine UDF dependencies."""
    udf_dependencies = set()
    raw_routines = read_routine_dir()
    udf_files = [file for file in artifact_files if file.name == UDF_FILE]
    for udf_file in udf_files:
        # all referenced UDFs need to be deployed in the same stage project due to access restrictions
        raw_routine = RawRoutine.from_file(udf_file)
        udfs_to_publish = accumulate_dependencies([], raw_routines, raw_routine.name)

        for dependency in udfs_to_publish:
            if dependency in raw_routines:
                file_path = Path(raw_routines[dependency].filepath)
                udf_dependencies.add(file_path)

    return udf_dependencies


def _view_dependencies(artifact_files, sql_dir):
    """Determine view dependencies."""
    view_dependencies = set()
    view_dependency_files = [file for file in artifact_files if file.name == VIEW_FILE]
    id_token = get_id_token()
    for dep_file in view_dependency_files:
        # all references views and tables need to be deployed in the same stage project
        if dep_file not in artifact_files:
            view_dependencies.add(dep_file)

        if dep_file.name == VIEW_FILE:
            view = View.from_file(dep_file, id_token=id_token)

            for dependency in view.table_references:
                dependency_components = dependency.split(".")
                if dependency_components[1:2] == ["INFORMATION_SCHEMA"]:
                    dependency_components.insert(0, view.project)
                if dependency_components[2:3] == ["INFORMATION_SCHEMA"]:
                    # INFORMATION_SCHEMA has more components that will be treated as the table name
                    # no deploys for INFORMATION_SCHEMA will happen later on
                    dependency_components = dependency_components[:2] + [
                        ".".join(dependency_components[2:])
                    ]
                if len(dependency_components) != 3:
                    raise ValueError(
                        f"Invalid table reference {dependency} in view {view.name}. "
                        "Tables should be fully qualified, expected format: project.dataset.table."
                    )
                project, dataset, name = dependency_components

                file_path = Path(view.path).parent.parent.parent / dataset / name

                file_exists_for_dependency = False
                for file in [VIEW_FILE, QUERY_FILE, QUERY_SCRIPT, MATERIALIZED_VIEW]:
                    if (file_path / file).is_file():
                        if (file_path / file) not in artifact_files:
                            view_dependency_files.append(file_path / file)

                        file_exists_for_dependency = True
                        break

                path = Path(sql_dir) / project / dataset / name
                if not path.exists():
                    path.mkdir(parents=True, exist_ok=True)
                    # don't create schema for wildcard and metadata tables
                    if "*" not in name and name != "INFORMATION_SCHEMA":
                        partitioned_by = "submission_timestamp"
                        schema = Schema.for_table(
                            project=project,
                            dataset=dataset,
                            table=name,
                            id_token=id_token,
                            partitioned_by=partitioned_by,
                        )
                        schema.to_yaml_file(path / SCHEMA_FILE)

                if not file_exists_for_dependency:
                    (path / QUERY_SCRIPT).write_text(
                        "# Table stub generated by stage deploy"
                    )
                    view_dependencies.add(path / QUERY_SCRIPT)

            # extract UDF references from view definition
            raw_routines = read_routine_dir()
            udf_dependencies = set()

            for udf_dependency in view.udf_references:
                routine = raw_routines[udf_dependency]
                udf_dependencies.add(Path(routine.filepath))

            # determine UDF dependencies recursively
            view_dependencies.update(_udf_dependencies(udf_dependencies))
            view_dependencies.update(udf_dependencies)

    return view_dependencies


def _update_references(artifact_files, project_id, dataset_suffix, sql_dir):
    replace_references = []
    replace_partial_references = []
    for artifact_file in artifact_files:
        name = artifact_file.parent.name
        name_pattern = name.replace("*", r"\*")  # match literal *
        original_dataset = artifact_file.parent.parent.name
        original_project = artifact_file.parent.parent.parent.name

        deployed_dataset = original_dataset

        if original_dataset not in (
            "INFORMATION_SCHEMA",
            "region-eu",
            "region-us",
        ):
            deployed_dataset += f"_{original_project.replace('-', '_')}"
            if dataset_suffix:
                deployed_dataset += f"_{dataset_suffix}"

        deployed_project = project_id

        # Replace references, preserving fully quoted references.
        replace_partial_references += [
            # partially qualified references (like "telemetry.main")
            (
                re.compile(rf"(?<![\._])`{original_dataset}\.{name_pattern}`"),
                f"`{deployed_project}.{deployed_dataset}.{name}`",
                original_project,
            ),
            (
                re.compile(
                    rf"(?<![\._])`?{original_dataset}`?\.`?{name_pattern}(?![a-zA-Z0-9_])`?"
                ),
                f"`{deployed_project}`.`{deployed_dataset}`.`{name}`",
                original_project,
            ),
        ]
        replace_references += [
            # fully qualified references (like "moz-fx-data-shared-prod.telemetry.main")
            (
                re.compile(
                    rf"`{original_project}\.{original_dataset}\.{name_pattern}`"
                ),
                f"`{deployed_project}.{deployed_dataset}.{name}`",
                original_project,
            ),
            (
                re.compile(
                    rf"(?<![a-zA-Z0-9_])`?{original_project}`?\.`?{original_dataset}`?\.`?{name_pattern}(?![a-zA-Z0-9_])`?"
                ),
                f"`{deployed_project}`.`{deployed_dataset}`.`{name}`",
                original_project,
            ),
        ]

    for path in map(Path, glob(f"{sql_dir}/**/*.sql", recursive=True)):
        # apply substitutions
        if path.is_file():
            if "is_init()" in path.read_text():
                init_sql = render(
                    path.name,
                    template_folder=path.parent,
                    format=False,
                    **{"is_init": lambda: True},
                )
                query_sql = render(
                    path.name,
                    template_folder=path.parent,
                    format=False,
                    **{"is_init": lambda: False},
                )
                sql = f"""
                    {{% if is_init() %}}
                    {init_sql}
                    {{% else %}}
                    {query_sql}
                    {{% endif %}}
                """
            else:
                sql = render(path.name, template_folder=path.parent, format=False)

            for ref in replace_references:
                sql = re.sub(ref[0], ref[1], sql)

            for ref in replace_partial_references:
                file_project = path.parent.parent.parent.name
                if file_project == ref[2]:
                    sql = re.sub(ref[0], ref[1], sql)

            path.write_text(sql)


def _deploy_artifacts(ctx, artifact_files, project_id, dataset_suffix, sql_dir):
    """Deploy routines, tables and views."""
    # give read permissions to dry run accounts
    dataset_access_entries = [
        bigquery.AccessEntry(
            role="READER",
            entity_type=EntityTypes.USER_BY_EMAIL,
            entity_id=dry_run_account,
        )
        for dry_run_account in ConfigLoader.get(
            "dry_run", "function_accounts", fallback=[]
        )
    ]

    # deploy routines
    routine_files = [file for file in artifact_files if file.name in ROUTINE_FILES]
    for routine_file in routine_files:
        dataset = routine_file.parent.parent.name
        create_dataset_if_not_exists(
            project_id=project_id,
            dataset=dataset,
            suffix=dataset_suffix,
            access_entries=dataset_access_entries,
        )
    ctx.invoke(publish_routine, name=None, project_id=project_id, dry_run=False)

    # deploy table schemas
    query_files = list(
        {
            file
            for file in artifact_files
            if file.name in [QUERY_FILE, QUERY_SCRIPT]
            # don't attempt to deploy wildcard or metadata tables
            and "*" not in file.parent.name and file.parent.name != "INFORMATION_SCHEMA"
        }
    )

    if len(query_files) > 0:
        # checking and creating datasets needs to happen sequentially
        for query_file in query_files:
            dataset = query_file.parent.parent.name
            create_dataset_if_not_exists(
                project_id=project_id,
                dataset=dataset,
                suffix=dataset_suffix,
                access_entries=dataset_access_entries,
            )

        ctx.invoke(
            update_query_schema,
            name=query_files,
            sql_dir=sql_dir,
            project_id=project_id,
            respect_dryrun_skip=True,
            is_init=True,
        )
        ctx.invoke(
            deploy_query_schema,
            name=query_files,
            sql_dir=sql_dir,
            project_id=project_id,
            force=True,
            respect_dryrun_skip=False,
            skip_external_data=True,
        )

    # deploy views
    view_files = [
        file
        for file in artifact_files
        if file.name == VIEW_FILE and str(file) not in DryRun.skipped_files()
    ]
    for view_file in view_files:
        dataset = view_file.parent.parent.name
        create_dataset_if_not_exists(
            project_id=project_id,
            dataset=dataset,
            suffix=dataset_suffix,
            access_entries=dataset_access_entries,
        )

    ctx.invoke(
        publish_view,
        name=None,
        sql_dir=sql_dir,
        project_id=project_id,
        dry_run=False,
        skip_authorized=False,
        force=True,
        respect_dryrun_skip=True,
    )


def create_dataset_if_not_exists(project_id, dataset, suffix=None, access_entries=None):
    """Create a temporary dataset if not already exists."""
    client = bigquery.Client(project_id)
    dataset = bigquery.Dataset(f"{project_id}.{dataset}")
    dataset.location = "US"
    dataset = client.create_dataset(dataset, exists_ok=True)
    dataset.default_table_expiration_ms = 60 * 60 * 1000 * 12  # ms

    # mark dataset as expired 12 hours from now; can be removed by CI
    expiration = int(
        ((datetime.utcnow() - datetime(1970, 1, 1)).total_seconds() + 60 * 60 * 12)
        * 1000
    )
    dataset.labels = {"expires_on": expiration}
    if suffix:
        dataset.labels["suffix"] = suffix
    if access_entries:
        dataset.access_entries = dataset.access_entries + access_entries

    return client.update_dataset(
        dataset, ["default_table_expiration_ms", "labels", "access_entries"]
    )


@stage.command(
    help="""
        Remove deployed artifacts from stage environment

    Examples:
    ./bqetl stage clean
    """
)
@click.option(
    "--project-id",
    "--project_id",
    help="GCP project to deploy artifacts to",
    default="bigquery-etl-integration-test",
)
@click.option(
    "--dataset-suffix",
    "--dataset_suffix",
    help="Suffix appended to the deployed dataset",
)
@click.option(
    "--delete-expired",
    "--delete_expired",
    help="Remove all datasets that have expired including those that do not match the suffix",
    default=False,
    is_flag=True,
)
def clean(project_id, dataset_suffix, delete_expired):
    """Reset the stage environment."""
    client = bigquery.Client(project_id)

    dataset_filter = (
        None if delete_expired is True else f"labels.suffix:{dataset_suffix}"
    )
    datasets = client.list_datasets(filter=dataset_filter)

    current_timestamp = int(
        (datetime.utcnow() - datetime(1970, 1, 1)).total_seconds() * 1000
    )

    for dataset in datasets:
        if dataset.labels:
            # remove datasets that either match the suffix or are expired
            for label, value in dataset.labels.items():
                if (
                    dataset_suffix and label == "suffix" and value == dataset_suffix
                ) or (
                    delete_expired
                    and label == "expires_on"
                    and int(value) < current_timestamp
                ):
                    click.echo(f"Deleting dataset {dataset.full_dataset_id}")
                    try:
                        client.delete_dataset(dataset, delete_contents=True)
                    except NotFound as e:
                        # account for other concurrent CI runs
                        click.echo(f"Failed to delete: {e.message}")



"""Docs."""

import logging
import os
import shutil
import sys
import tempfile
from pathlib import Path

import rich_click as click

from bigquery_etl.config import ConfigLoader
from bigquery_etl.dryrun import DryRun

EXAMPLE_DIR = "examples"
INDEX_MD = "index.md"


@click.group("docs", help="Commands for generated documentation.")
def docs_():
    """Create the CLI group for doc commands."""
    pass


project_dirs_option = click.option(
    "--project_dirs",
    "--project-dirs",
    help="Directories of projects documentation is generated for.",
    multiple=True,
    default=[
        ConfigLoader.get("default", "sql_dir") + "/" + project + "/"
        for project in ConfigLoader.get("docs", "default_projects")
    ],
)

log_level_option = click.option(
    "--log-level",
    "--log_level",
    help="Log level.",
    default=logging.getLevelName(logging.INFO),
    type=str.upper,
)


@docs_.command("generate", help="Generate the project docs.")
@project_dirs_option
@click.option(
    "--docs_dir",
    "--docs-dir",
    default=ConfigLoader.get("docs", "docs_dir"),
    help="Directory containing static documentation.",
)
@click.option(
    "--output_dir",
    "--output-dir",
    required=True,
    help="Generated documentation is written to this output directory.",
)
@log_level_option
def generate(project_dirs, docs_dir, output_dir, log_level):
    """Generate documentation for project."""
    from bigquery_etl.docs.bqetl.generate_bqetl_docs import generate_bqetl_docs
    from bigquery_etl.docs.mozfun.generate_mozfun_docs import generate_udf_docs

    out_dir = os.path.join(output_dir, "docs")

    # To customize Mkdocs, we need to extend the theme with an `overrides` folder
    # https://squidfunk.github.io/mkdocs-material/customization/#overriding-partials
    override_dir = os.path.join(output_dir, "overrides")

    if os.path.exists(out_dir) and os.path.exists(override_dir):
        shutil.rmtree(out_dir)
        shutil.rmtree(override_dir)

    # copy assets from /docs and /overrides folders to output folder
    shutil.copytree(docs_dir, out_dir)
    shutil.copytree("bigquery_etl/docs/overrides", override_dir)

    # move mkdocs.yml out of docs/
    mkdocs_path = os.path.join(output_dir, "mkdocs.yml")
    shutil.move(os.path.join(out_dir, "mkdocs.yml"), mkdocs_path)

    # generate bqetl command docs
    generate_bqetl_docs(Path(out_dir) / "bqetl.md")

    # generate docs
    for project_dir in project_dirs:
        generate_udf_docs(out_dir, project_dir)


@docs_.command("validate", help="Validate the project docs.")
@project_dirs_option
@log_level_option
def validate(project_dirs, log_level):
    """Validate UDF docs."""
    from bigquery_etl.routine.parse_routine import read_routine_dir, sub_local_routines

    is_valid = True

    for project_dir in project_dirs:
        if os.path.isdir(project_dir):
            parsed_routines = read_routine_dir(project_dir)

            for root, dirs, files in os.walk(project_dir, followlinks=True):
                if os.path.basename(root) == EXAMPLE_DIR:
                    sql_files = (f for f in files if os.path.splitext(f)[1] == ".sql")
                    for file in sql_files:
                        dry_run_sql = sub_local_routines(
                            (Path(root) / file).read_text(),
                            project_dir,
                            parsed_routines,
                        )

                        # store sql in temporary file for dry_run
                        tmp_dir = Path(tempfile.mkdtemp()) / Path(root)
                        tmp_dir.mkdir(parents=True, exist_ok=True)
                        tmp_example_file = tmp_dir / file
                        tmp_example_file.write_text(dry_run_sql)

                        if not DryRun(str(tmp_example_file)).is_valid():
                            is_valid = False

    if not is_valid:
        print("Invalid examples.")
        sys.exit(1)



#!/usr/bin/env python3
"""clients_daily_histogram_aggregates query generator."""
import argparse
import json
import sys
import textwrap
import urllib.request
from pathlib import Path
from time import sleep

from google.cloud import bigquery

from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util import probe_filters

sys.path.append(str(Path(__file__).parent.parent.parent.resolve()))

PROBE_INFO_SERVICE = (
    "https://probeinfo.telemetry.mozilla.org/firefox/all/main/all_probes"
)

p = argparse.ArgumentParser()
p.add_argument(
    "--agg-type",
    type=str,
    help="One of histograms/keyed_histograms",
    required=True,
)
p.add_argument(
    "--json-output",
    action="store_true",
    help="Output the result wrapped in json parseable as an XCOM",
)
p.add_argument(
    "--wait-seconds",
    type=int,
    default=0,
    help="Add a delay before executing the script to allow time for the xcom sidecar to complete startup",
)
p.add_argument(
    "--processes",
    type=str,
    nargs="*",
    help="Processes to include in the output.  Defaults to all processes.",
)


def generate_sql(opts, additional_queries, windowed_clause, select_clause, json_output):
    """Create a SQL query for the clients_daily_histogram_aggregates dataset."""
    query = textwrap.dedent(
        f"""-- Query generated by: sql/telemetry_derived/clients_daily_histogram_aggregates_v1.sql.py --agg-type {opts["agg_type"]}
        CREATE TEMP FUNCTION udf_aggregate_json_sum(histograms ARRAY<STRING>) AS (
          ARRAY(
            SELECT AS STRUCT
              FORMAT('%d', values_entry.key) AS key,
              SUM(values_entry.value) AS value
            FROM
              UNNEST(histograms) AS histogram,
              UNNEST(mozfun.hist.extract(histogram).values) AS values_entry
            WHERE
              histogram IS NOT NULL
            GROUP BY
              values_entry.key
            ORDER BY
              values_entry.key
          )
        );

        WITH valid_build_ids AS (
            SELECT
              DISTINCT(build.build.id) AS build_id
            FROM
              `moz-fx-data-shared-prod.telemetry.buildhub2`
        ),
        filtered AS (
            SELECT
                *,
                SPLIT(application.version, '.')[OFFSET(0)] AS app_version,
                DATE(submission_timestamp) as submission_date,
                normalized_os as os,
                application.build_id AS app_build_id,
                normalized_channel AS channel
            FROM `moz-fx-data-shared-prod.telemetry_stable.main_v5`
            INNER JOIN valid_build_ids
            ON (application.build_id = build_id)
            WHERE DATE(submission_timestamp) = @submission_date
                AND normalized_channel in (
                  "release", "beta", "nightly"
                )
                AND client_id IS NOT NULL
        ),
        sampled_data AS (
          SELECT
            *
          FROM
            filtered
          WHERE
            channel IN ("nightly", "beta")
            OR (channel = "release" AND os != "Windows")
            OR (
                channel = "release" AND
                os = "Windows" AND
                MOD(sample_id, @sample_size) = 0)
        ),

        {additional_queries}

        aggregated AS (
            {windowed_clause}
        )
        {select_clause}
        """
    )

    if json_output:
        return json.dumps(query)
    else:
        return query


def _get_keyed_histogram_sql(probes_and_buckets):
    probes = probes_and_buckets["probes"]
    buckets = probes_and_buckets["buckets"]

    probes_struct = []
    for probe, details in probes.items():
        for process in details["processes"]:
            probe_location = (
                f"payload.keyed_histograms.{probe}"
                if process == "parent"
                else f"payload.processes.{process}.keyed_histograms.{probe}"
            )
            buckets_for_probe = "{min}, {max}, {num}".format(
                min=buckets[probe]["min"],
                max=buckets[probe]["max"],
                num=buckets[probe]["n_buckets"],
            )

            agg_string = (
                f"('{probe}', "
                f"'histogram-{details['type']}', "
                f"'{process}', "
                f"{probe_location}, "
                f"({buckets_for_probe}))"
            )

            probes_struct.append(agg_string)

    probes_struct.sort()
    probes_arr = ",\n\t\t\t".join(probes_struct)

    probes_string = """
        metric,
        metric_type,
        key,
        ARRAY_AGG(bucket_range) as bucket_range,
        ARRAY_AGG(value) as value
    """

    additional_queries = f"""
        grouped_metrics AS
          (SELECT
            DATE(submission_timestamp) as submission_date,
            sample_id,
            client_id,
            profile_group_id,
            normalized_os as os,
            SPLIT(application.version, '.')[OFFSET(0)] AS app_version,
            application.build_id AS app_build_id,
            normalized_channel AS channel,
            ARRAY<STRUCT<
                name STRING,
                metric_type STRING,
                process STRING,
                value ARRAY<STRUCT<key STRING, value STRING>>,
                bucket_range STRUCT<first_bucket INT64, last_bucket INT64, num_buckets INT64>
            >>[
              {probes_arr}
            ] as metrics
          FROM sampled_data),

          flattened_metrics AS
            (SELECT
              submission_date,
              sample_id,
              client_id,
              profile_group_id,
              os,
              app_version,
              app_build_id,
              channel,
              process,
              metrics.name AS metric,
              metrics.metric_type AS metric_type,
              bucket_range,
              value.key AS key,
              value.value AS value
            FROM grouped_metrics
            CROSS JOIN UNNEST(metrics) AS metrics
            CROSS JOIN unnest(metrics.value) AS value),
    """

    windowed_clause = f"""
        SELECT
            submission_date,
            sample_id,
            client_id,
            profile_group_id,
            os,
            app_version,
            app_build_id,
            channel,
            process,
            {probes_string}
            FROM flattened_metrics
            GROUP BY
                sample_id,
                client_id,
                profile_group_id,
                submission_date,
                os,
                app_version,
                app_build_id,
                channel,
                process,
                metric,
                metric_type,
                key
    """

    select_clause = """
        SELECT
            submission_date,
            sample_id,
            client_id,
            profile_group_id,
            os,
            app_version,
            app_build_id,
            channel,
            ARRAY_AGG(STRUCT<
                metric STRING,
                metric_type STRING,
                key STRING,
                process STRING,
                agg_type STRING,
                bucket_range STRUCT<
                    first_bucket INT64,
                    last_bucket INT64,
                    num_buckets INT64
                >,
                value ARRAY<STRUCT<key STRING, value INT64>>
            >(
                metric,
                metric_type,
                key,
                process,
                'summed_histogram',
                bucket_range[OFFSET(0)],
                udf_aggregate_json_sum(value)
            )) AS histogram_aggregates
        FROM aggregated
        GROUP BY
            sample_id,
            client_id,
            profile_group_id,
            submission_date,
            os,
            app_version,
            app_build_id,
            channel
    """

    return {
        "additional_queries": additional_queries,
        "select_clause": select_clause,
        "windowed_clause": windowed_clause,
    }


def get_histogram_probes_sql_strings(probes_and_buckets, histogram_type):
    """Put together the subsets of SQL required to query histograms."""
    probes = probes_and_buckets["probes"]
    buckets = probes_and_buckets["buckets"]

    sql_strings = {}
    if histogram_type == "keyed_histograms":
        return _get_keyed_histogram_sql(probes_and_buckets)

    probe_structs = []
    for probe, details in probes.items():
        for process in details["processes"]:
            probe_location = (
                f"payload.histograms.{probe}"
                if process == "parent"
                else f"payload.processes.{process}.histograms.{probe}"
            )
            buckets_for_probe = "{min}, {max}, {num}".format(
                min=buckets[probe]["min"],
                max=buckets[probe]["max"],
                num=buckets[probe]["n_buckets"],
            )

            agg_string = (
                f"('{probe}', "
                f"'histogram-{details['type']}', "
                f"'{process}', "
                f"{probe_location}, "
                f"({buckets_for_probe}))"
            )

            probe_structs.append(agg_string)

    probe_structs.sort()
    probes_arr = ",\n\t\t\t".join(probe_structs)
    probes_string = f"""
            ARRAY<STRUCT<
                metric STRING,
                metric_type STRING,
                process STRING,
                value STRING,
                bucket_range STRUCT<first_bucket INT64, last_bucket INT64, num_buckets INT64>
            >> [
            {probes_arr}
        ] AS histogram_aggregates
    """

    sql_strings[
        "select_clause"
    ] = """
        SELECT
          sample_id,
          client_id,
          profile_group_id,
          submission_date,
          os,
          app_version,
          app_build_id,
          channel,
          ARRAY_AGG(STRUCT<
            metric STRING,
            metric_type STRING,
            key STRING,
            process STRING,
            agg_type STRING,
            bucket_range STRUCT<first_bucket INT64, last_bucket INT64, num_buckets INT64>,
            value ARRAY<STRUCT<key STRING, value INT64>>
          > (metric,
            metric_type,
            '',
            process,
            'summed_histogram',
            bucket_range[OFFSET(0)],
            udf_aggregate_json_sum(value))) AS histogram_aggregates
        FROM aggregated
        GROUP BY
          1, 2, 3, 4, 5, 6, 7, 8

    """

    sql_strings[
        "additional_queries"
    ] = f"""
        histograms AS (
            SELECT
                submission_date,
                sample_id,
                client_id,
                profile_group_id,
                os,
                app_version,
                app_build_id,
                channel,
                {probes_string}
            FROM sampled_data),

        filtered_aggregates AS (
          SELECT
            submission_date,
            sample_id,
            client_id,
            profile_group_id,
            os,
            app_version,
            app_build_id,
            channel,
            metric,
            metric_type,
            process,
            bucket_range,
            value
          FROM histograms
          CROSS JOIN
            UNNEST(histogram_aggregates)
          WHERE value IS NOT NULL
        ),
    """

    sql_strings[
        "windowed_clause"
    ] = """
      SELECT
        sample_id,
        client_id,
        profile_group_id,
        submission_date,
        os,
        app_version,
        app_build_id,
        channel,
        metric,
        metric_type,
        process,
        ARRAY_AGG(bucket_range) AS bucket_range,
        ARRAY_AGG(value) AS value
      FROM filtered_aggregates
      GROUP BY
        1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11
    """

    return sql_strings


def get_histogram_probes_and_buckets(histogram_type, processes_to_output):
    """Return relevant histogram probes."""
    project = "moz-fx-data-shared-prod"
    main_summary_histograms = {}

    client = bigquery.Client(project)
    table = client.get_table("telemetry_stable.main_v5")
    main_summary_schema = [field.to_api_repr() for field in table.schema]

    # Fetch the histograms field
    histograms_field = []
    for field in main_summary_schema:
        if field["name"] != "payload":
            continue

        for payload_field in field["fields"]:
            if payload_field["name"] == histogram_type:
                histograms_field.append(
                    {"histograms": payload_field, "process": "parent"}
                )
                continue

            if payload_field["name"] == "processes":
                for processes_field in payload_field["fields"]:
                    if processes_field["name"] in ["content", "gpu"]:
                        process_field = processes_field["name"]
                        for type_field in processes_field["fields"]:
                            if type_field["name"] == histogram_type:
                                histograms_field.append(
                                    {"histograms": type_field, "process": process_field}
                                )
                                break

    if len(histograms_field) == 0:
        return

    for histograms_and_process in histograms_field:
        for histogram in histograms_and_process["histograms"].get("fields", {}):
            if "name" not in histogram:
                continue

            processes = main_summary_histograms.setdefault(histogram["name"], set())
            if (
                processes_to_output is None
                or histograms_and_process["process"] in processes_to_output
            ):
                processes.add(histograms_and_process["process"])
            main_summary_histograms[histogram["name"]] = processes

    with urllib.request.urlopen(PROBE_INFO_SERVICE) as url:
        data = json.loads(url.read())
        excluded_probes = probe_filters.get_etl_excluded_probes_quickfix("desktop")
        histogram_probes = {
            x.replace("histogram/", "").replace(".", "_").lower()
            for x in data.keys()
            if x.startswith("histogram/")
        }

        bucket_details = {}
        relevant_probes = {
            histogram: {"processes": process}
            for histogram, process in main_summary_histograms.items()
            if histogram in histogram_probes and histogram not in excluded_probes
        }
        for key in data.keys():
            if not key.startswith("histogram/"):
                continue

            channel = "nightly"
            if "nightly" not in data[key]["history"]:
                channel = "beta"

                if "beta" not in data[key]["history"]:
                    channel = "release"

            data_details = data[key]["history"][channel][0]["details"]
            probe = key.replace("histogram/", "").replace(".", "_").lower()

            # Some keyed GPU metrics aren't correctly flagged as type
            # "keyed_histograms", so we filter those out here.
            if processes_to_output is None or "gpu" in processes_to_output:
                if data_details["keyed"] == (histogram_type == "histograms"):
                    try:
                        del relevant_probes[probe]
                    except KeyError:
                        pass
                    continue

            if probe in relevant_probes:
                relevant_probes[probe]["type"] = data_details["kind"]

            # NOTE: some probes, (e.g. POPUP_NOTIFICATION_MAINACTION_TRIGGERED_MS) have values
            # in the probe info service like 80 * 25 for the value of n_buckets.
            # So they do need to be evaluated as expressions.
            bucket_details[probe] = {
                "n_buckets": int(eval(str(data_details["n_buckets"]))),
                "min": int(eval(str(data_details["low"]))),
                "max": int(eval(str(data_details["high"]))),
            }

        return {"probes": relevant_probes, "buckets": bucket_details}


def main(argv, out=print):
    """Print a clients_daily_histogram_aggregates query to stdout."""
    opts = vars(p.parse_args(argv[1:]))
    sql_string = ""

    if opts["agg_type"] in ("histograms", "keyed_histograms"):
        probes_and_buckets = get_histogram_probes_and_buckets(
            opts["agg_type"], opts["processes"]
        )
        sql_string = get_histogram_probes_sql_strings(
            probes_and_buckets, opts["agg_type"]
        )
    else:
        raise ValueError("agg-type must be one of histograms, keyed_histograms")

    sleep(opts["wait_seconds"])
    out(
        reformat(
            generate_sql(
                opts,
                sql_string.get("additional_queries", ""),
                sql_string["windowed_clause"],
                sql_string["select_clause"],
                opts["json_output"],
            )
        )
    )


if __name__ == "__main__":
    main(sys.argv)
